from spaceone.core.auth import Authenticator
from spaceone.core.auth.jwt.jwt_util import JWTUtil
from spaceone.core.error import *


class JWTAuthenticator(Authenticator):

    def __init__(self, key):
        self._key = key

    def validate(self, token, options=None):
        if not self._key:
            raise ERROR_AUTHENTICATE_FAILURE(message='Decode key is not set.')

        if not isinstance(token, (str, bytes)):
            raise ERROR_AUTHENTICATE_FAILURE(message='Invalid token format.')

        if options is None:
            options = {}

        try:
            payload = JWTUtil.decode(token, self._key, options=options)
        except Exception:
            raise ERROR_AUTHENTICATE_FAILURE(message='Token is invalid or expired.')

        return payload



from spaceone.core.service.utils import *
from spaceone.core.service.service import *

__all__ = ['BaseService', 'transaction', 'authentication_handler', 'authorization_handler', 'mutation_handler',
           'event_handler', 'check_required', 'append_query_filter', 'change_tag_filter', 'change_timestamp_value',
           'change_timestamp_filter', 'change_date_value', 'append_keyword_filter', 'change_only_key',
           'set_query_page_limit']



import base64
import json

from jwcrypto import jwk
from jwcrypto import jwt as jwcrypto_jwt


class JWTUtil:
    @staticmethod
    def generate_jwk(key_type="RSA", size=2048):
        key = jwk.JWK.generate(kty=key_type, size=size)
        private_jwk = json.loads(key.export_private())
        public_jwk = json.loads(key.export_public())
        return private_jwk, public_jwk

    @staticmethod
    def encode(payload: dict, private_jwk: dict, algorithm="RS256") -> str:
        # Convert dict to JWK object
        key = jwk.JWK(**private_jwk)

        # Create JWT object with claims and header
        jwt_obj = jwcrypto_jwt.JWT(claims=payload, header={"alg": algorithm})

        # Sign the token
        jwt_obj.make_signed_token(key)

        # Serialize to compact format
        return jwt_obj.serialize()

    @staticmethod
    def decode(token: str, public_jwk: dict, algorithm="RS256", options=None) -> dict:
        if options is None:
            options = {}

        # Convert dict to JWK object
        key = jwk.JWK(**public_jwk)

        # Create JWT object and deserialize
        jwt_obj = jwcrypto_jwt.JWT(jwt=token, key=key, algs=[algorithm])

        # Validate the token
        verify_aud = options.get("verify_aud", False)
        check_claims = None
        if verify_aud and "aud" in options:
            check_claims = {"aud": options["aud"]}

        if check_claims:
            jwt_obj._check_claims = check_claims

        jwt_obj.validate(key)

        # Parse claims from JSON string
        return json.loads(jwt_obj.claims)

    @staticmethod
    def unverified_decode(token: str) -> dict:
        parts = token.split(".")
        if len(parts) != 3:
            raise ValueError("Invalid JWT token format")

        # Decode payload part (base64url)
        payload_part = parts[1]

        # Handle base64url padding
        padding = 4 - len(payload_part) % 4
        if padding != 4:
            payload_part += "=" * padding

        # Decode base64url
        payload_bytes = base64.urlsafe_b64decode(payload_part)
        payload_str = payload_bytes.decode("utf-8")

        # Parse JSON
        return json.loads(payload_str)



import types
import logging
from google.protobuf.json_format import MessageToDict
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from spaceone.core import config as global_config
from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_grpc_endpoint
from spaceone.core.error import *

__all__ = ["SpaceConnector"]

_LOGGER = logging.getLogger(__name__)


class SpaceConnector(BaseConnector):

    def __init__(self, return_type: str = 'dict', config: dict = None,
                 service: str = None, endpoint: str = None, **kwargs):
        super().__init__(config)

        self._mock_mode = global_config.get_global('MOCK_MODE', False)
        self._service = service
        self._endpoint = endpoint
        self._return_type = return_type
        self._token = kwargs.get('token')
        self._endpoints = self.config.get('endpoints', {})
        self._verify()

        if self._mock_mode is False:
            self._init_client()

    @property
    def client(self):
        return self._client

    def dispatch(self, method: str, params: dict = None, **kwargs):
        return self._call_api(method, params, **kwargs)

    def _call_api(self, method: str, params: dict = None, **kwargs):
        token = kwargs.get('token')

        self._check_mock_mode(method)

        resource, verb = self._parse_method(method)
        self._check_method(resource, verb)

        params = params or {}
        kwargs['metadata'] = self._get_connection_metadata(token)

        response_or_iterator = getattr(getattr(self._client, resource), verb)(params, **kwargs)

        if self._return_type == 'dict':
            if isinstance(response_or_iterator, types.GeneratorType):
                return self._generate_response(response_or_iterator)
            else:
                return self._change_message(response_or_iterator)
        else:
            return response_or_iterator

    def _check_mock_mode(self, method):
        if self._mock_mode:
            raise ERROR_CONNECTOR(connector='SpaceConnector',
                                  reason=f'Dispatch cannot be executed in mock mode. '
                                         f'(endpoint = {self._get_endpoint()}, method = {method})')

    def _verify(self):
        if self._endpoint:
            pass
        elif self._service:
            if not isinstance(self._endpoints, dict):
                raise ERROR_CONNECTOR_CONFIGURATION(connector='SpaceConnector')

            if self._service not in self._endpoints:
                raise ERROR_CONNECTOR_LOAD(connector='SpaceConnector', reason=f'{self._service} endpoint is undefined.')
        else:
            raise ERROR_CONNECTOR_LOAD(connector='SpaceConnector', reason='service or endpoint is required.')

    def _init_client(self):
        endpoint = self._get_endpoint()
        e = parse_grpc_endpoint(endpoint)
        self._client = pygrpc.client(endpoint=e['endpoint'], ssl_enabled=e['ssl_enabled'],
                                     max_message_length=1024 * 1024 * 256)

    @staticmethod
    def _change_message(message):
        return MessageToDict(message, preserving_proto_field_name=True)

    def _generate_response(self, response_iterator):
        for response in response_iterator:
            yield self._change_message(response)

    def _get_connection_metadata(self, token=None):
        metadata = []

        if token:
            metadata.append(('token', token))
        elif self._token:
            metadata.append(('token', self._token))
        elif token := self.transaction.meta.get('token'):
            metadata.append(('token', token))

        carrier = {}
        TraceContextTextMapPropagator().inject(carrier)

        if traceparent := carrier.get('traceparent'):
            metadata.append(('traceparent', traceparent))

        return metadata

    def _parse_method(self, method):
        try:
            resource, verb = method.split('.')
        except Exception:
            raise ERROR_CONNECTOR(connector='SpaceConnector',
                                  reason=f'Method is invalid. (endpoint = {self._get_endpoint()}, method = {method})')

        return resource, verb

    def _check_method(self, resource, verb):
        supported_verb = self._client.api_resources.get(resource)

        if supported_verb is None or verb not in supported_verb:
            raise ERROR_CONNECTOR(connector='SpaceConnector',
                                  reason=f'Method not supported. (endpoint = {self._get_endpoint()}, '
                                         f'method = {resource}.{verb})')

    def _get_endpoint(self):
        return self._endpoint or self._endpoints[self._service]



import logging
import time
from concurrent import futures

import grpc
from spaceone.core import config
from spaceone.core.logger import set_logger
from spaceone.core.opentelemetry import set_tracer, set_metric
from grpc_reflection.v1alpha import reflection

_LOGGER = logging.getLogger(__name__)


class _ServerInterceptor(grpc.ServerInterceptor):
    _SKIP_METHODS = (
        '/grpc.reflection.v1alpha.ServerReflection/ServerReflectionInfo'
    )

    def _check_skip_method(self, method):
        is_skip = False

        if method in self._SKIP_METHODS:
            is_skip = True

        return is_skip

    def intercept_service(self, continuation, handler_call_details):
        # is_skip = self._check_skip_method(handler_call_details.method)
        #
        # if not is_skip:
        #     pass

        response = continuation(handler_call_details)

        # if not is_skip:
        #     pass

        return response


def _get_proto_conf(package):
    proto_module = __import__(f'{package}.conf.proto_conf', fromlist=['proto_conf'])
    return getattr(proto_module, 'PROTO', {})


def _init_services(server):
    service_names = []

    # Set Core Services
    proto_conf = _get_proto_conf(config.get_package())
    server, service_names = _add_services(server, service_names, proto_conf)

    # Set Extension Services
    proto_conf = config.get_global('GRPC_EXTENSION_APIS', {})
    server, service_names = _add_services(server, service_names, proto_conf)

    return server, service_names


def _add_services(server, service_names: list, proto_conf: dict):
    for module_path, servicer_names in proto_conf.items():
        for servicer_name in servicer_names:
            api_module = _import_module(module_path, servicer_name)
            if api_module:
                if hasattr(api_module, servicer_name):
                    servicer = getattr(api_module, servicer_name)()

                    getattr(servicer.pb2_grpc_module, f'add_{servicer.name}Servicer_to_server')(servicer, server)
                    service_names.append(servicer.service_name)
                else:
                    _LOGGER.warning(f'[_add_services] Failed to add service. '
                                    f'(module_path={module_path}, servicer_name={servicer_name})')

    return server, service_names


def _import_module(module_path, servicer_name):
    module = None
    try:
        module = __import__(module_path, fromlist=[servicer_name])
    except Exception as e:
        _LOGGER.warning(f'[_import_module] Cannot import grpc servicer module. (reason = {e})', exc_info=True)

    return module


def _get_grpc_options(conf):
    options = []
    if 'MAX_MESSAGE_LENGTH' in conf:
        options += [
            ('grpc.max_send_message_length', conf['MAX_MESSAGE_LENGTH']),
            ('grpc.max_receive_message_length', conf['MAX_MESSAGE_LENGTH']),
        ]

    return options


def serve():
    conf = config.get_global()

    # Enable logging configuration
    set_logger()

    # Set OTel Tracer and Metric
    set_tracer()

    server_interceptor = _ServerInterceptor()
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=conf['MAX_WORKERS']),
        interceptors=(server_interceptor,),
        # options=_get_grpc_options(conf)
    )

    server, service_names = _init_services(server)

    service_names_str = '\n\t - '.join(service_names)
    _LOGGER.debug(f'Loaded Services: \n\t - {service_names_str}')
    reflection.enable_server_reflection(service_names, server)

    server.add_insecure_port(f'[::]:{conf["PORT"]}')
    _LOGGER.info(f'Start gRPC Server ({config.get_service()}): '
                 f'port={conf["PORT"]}, max_workers={conf["MAX_WORKERS"]}')
    server.start()
    server.wait_for_termination()



import json
import random
import string
import logging

from multiprocessing import Process

from spaceone.core import queue, config
from spaceone.core.locator import Locator
from spaceone.core.logger import set_logger
from spaceone.core.error import ERROR_TASK_LOCATOR, ERROR_TASK_METHOD

_LOGGER = logging.getLogger(__name__)


def randomString(stringLength=8):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.sample(letters, stringLength))


class SingleTask:

    def __init__(self, single_task: dict):
        self.locator = single_task['locator']
        self.name = single_task['name']
        self.metadata = single_task['metadata']
        self.method = single_task['method']
        self.params = single_task['params']
        self._locator = Locator()

    def execute(self):
        """ Run method
        """
        try:
            if self.locator == 'SERVICE':
                caller = self._locator.get_service(self.name, self.metadata)
            elif self.locator == 'MANAGER':
                caller = self._locator.get_manager(self.name)

        except Exception as e:
            _LOGGER.debug(f'[SingleTask] fail at locator {e}')
            raise ERROR_TASK_LOCATOR(locator=self.locator, name=self.name)

        try:
            _LOGGER.debug(f'[SingleTask] request: {self.name}.{self.method} {self.params}')
            method = getattr(caller, self.method)
            resp = method(**self.params)
            _LOGGER.debug(f'[SingleTask] response: {resp}')
            return resp
        except Exception as e:
            _LOGGER.error(f'[SingleTask] fail to execute method: {self.method}, params: {self.params}, {e}')
            raise ERROR_TASK_METHOD(name=self.name, method=self.method, params=self.params)


class SpaceoneTask:
    def __init__(self, task: dict):
        self.task = task
        self.name = task.get('name', "")
        self.version = task.get('version', 'v1')
        self.execution_engine = task.get('executionEngine', 'BaseWorker')
        self.stages = task.get('stages')
        self.stop_on_failure = task.get('stop_on_failure', True)

    def execute(self):
        """ Run stage
        """
        self.tasks = []
        for stage in self.stages:
            try:
                single_task = SingleTask(stage)
                single_task.execute()
            except Exception as e:
                _LOGGER.error(f'[SpaceoneTask] fail to parse {stage}, {e}')
                if self.stop_on_failure:
                    _LOGGER.debug(f'[SpaceoneTask] stop task, since stop on failure is enabled')
                    break


class BaseWorker(Process):

    def __init__(self, queue, **kwargs):
        self._name_ = 'worker-%s' % randomString()
        self.queue = queue
        _LOGGER.debug(f'[BaseWorker] BaseWorker name  : {self._name_}')
        _LOGGER.debug(f'[BaseWorker] BaseWorker queue : {self.queue}')

        self.global_config = config.get_global()
        super().__init__()

    def run(self):
        """ Infinite Loop
        """
        config.set_global_force(**self.global_config)

        # Enable logging configuration
        set_logger()

        while True:
            # Read from Queue
            binary_task = queue.get(self.queue)
            try:
                json_task = json.loads(binary_task.decode())
                task = SpaceoneTask(json_task)
                # Run task
                task.execute()

            except Exception as e:
                _LOGGER.error(f'[{self._name_}] failed to decode task: {binary_task}, {e}')
                continue





import copy
import logging
import consul

from spaceone.core import utils
from spaceone.core.config import default_conf

_REMOTE_URL = []
_GLOBAL = {}
_LOGGER = logging.getLogger(__name__)


def init_conf(package, **kwargs):
    set_default_conf()

    _GLOBAL['PACKAGE'] = package
    _GLOBAL['SERVICE'] = package.rsplit('.', 1)[-1:][0]

    if 'host' in kwargs:
        _GLOBAL['HOST'] = kwargs['HOST']

    if 'port' in kwargs:
        _GLOBAL['PORT'] = kwargs['port']


def set_default_conf():
    for key, value in vars(default_conf).items():
        if not key.startswith('__'):
            _GLOBAL[key] = value


def get_package():
    return _GLOBAL['PACKAGE']


def get_service():
    return _GLOBAL['SERVICE']


def get_handler(name):
    return _GLOBAL.get('HANDLERS', {}).get(name, {})


def get_connector(name):
    return _GLOBAL.get('CONNECTORS', {}).get(name, {})


def set_service_config():
    """
    Get config from service ({package}.conf.global_conf)
    """

    package = _GLOBAL['PACKAGE']
    if package is None:
        raise ValueError(f'Package is undefined.')

    global_module = __import__(f'{package}.conf.global_conf', fromlist=['global_conf'])
    for key, value in vars(global_module).items():
        if not key.startswith('__'):
            _GLOBAL[key] = value


def get_global(key=None, default=None):
    if key:
        return copy.deepcopy(_GLOBAL.get(key, default))
    else:
        return copy.deepcopy(_GLOBAL)


def set_global(**config):
    global_conf = get_global()

    for key, value in config.items():
        if key in global_conf:
            if not isinstance(value, type(global_conf[key])) and global_conf[key] is not None:
                value_type_name = type(global_conf[key]).__name__
                raise ValueError(f'Value type is invalid. (GLOBAL.{key} = {value_type_name})')

            if isinstance(value, dict):
                global_conf[key] = utils.deep_merge(value, global_conf[key])
            else:
                global_conf[key] = value

    _GLOBAL.update(global_conf)


def set_global_force(**config):
    for key, value in config.items():
        _GLOBAL[key] = value


def set_file_conf(config_yml: str):
    file_conf: dict = utils.load_yaml_from_file(config_yml)
    global_conf: dict = file_conf.get('GLOBAL', {})
    set_global(**global_conf)

    import_conf: list = file_conf.get('IMPORT', [])
    if isinstance(import_conf, list):
        for uri in import_conf:
            import_remote_conf(uri)

    # DEPRECATED: REMOTE_URL setting changed to IMPORT
    import_conf: list = file_conf.get('REMOTE_URL', [])
    if isinstance(import_conf, list):
        for uri in import_conf:
            import_remote_conf(uri)


def import_remote_conf(uri):
    endpoint = utils.parse_endpoint(uri)
    scheme = endpoint.get('scheme')

    remote_conf = None

    if scheme == 'file':
        remote_conf = utils.load_yaml_from_file(endpoint['path'])

    elif scheme in ['http', 'https']:
        remote_conf = utils.load_yaml_from_url(uri)

    elif scheme == 'consul':
        remote_conf = load_consul_config(endpoint)

    if isinstance(remote_conf, dict):
        set_global(**remote_conf)


def load_consul_config(endpoint):
    hostname = endpoint.get('hostname')
    port = endpoint.get('port')
    key = endpoint.get('path', '')[1:]

    try:
        conf = {}
        if hostname:
            conf['host'] = hostname

        if port:
            conf['port'] = port

        c = consul.Consul(**conf)
        index, data = c.kv.get(key)
        if data:
            print(data)
            json_str = data['Value'].decode('utf-8')
            return utils.load_json(json_str)

        return {}
    except Exception as e:
        raise Exception(f'Consul Call Error: {e}')



from spaceone.core.error import *
from spaceone.core import utils

__all__ = ['STAT_GROUP_OPERATORS', 'STAT_PROJECT_OPERATORS']


def _group_count_resolver(condition, key, operator, name, data_type, sub_conditions, *args):
    if sub_conditions:
        return {
            name: {
                '$sum': {
                    '$cond': [
                        sub_conditions,
                        1,
                        0
                    ]
                }
            }
        }
    else:
        return {
            name: {'$sum': 1}
        }


def _group_push_resolver(condition, key, operator, name, data_type, sub_conditions, sub_fields, datetime_fields, *args):
    if not key and len(sub_fields) == 0:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires fields: {condition}")

    if sub_conditions:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition's conditions not supported: {condition}")

    push_query = {}

    if key:
        return {
            name: {
                '$push': f'${key}'
            }
        }
    else:
        for sub_field in sub_fields:
            f_key = sub_field.get('key', sub_field.get('k'))
            f_name = sub_field.get('name', sub_field.get('n'))

            if f_key in datetime_fields:
                push_query[f_name] = {
                    '$dateToString': {
                        'format': '%Y-%m-%dT%H:%M:%SZ',
                        'date': f'${f_key}'
                    }
                }
            else:
                push_query[f_name] = f'${f_key}'


        return {
            name: {
                '$push': push_query
            }
        }


def _group_average_resolver(condition, key, operator, name, data_type, sub_conditions, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires a key: {condition}")

    if sub_conditions:
        return {
            name: {
                '$avg': {
                    '$cond': [
                        sub_conditions,
                        f'${key}',
                        0
                    ]
                }
            }
        }
    elif data_type == 'array':
        return {
            name: {'$avg': {'$avg': f'${key}'}}
        }
    else:
        return {
            name: {'$avg': f'${key}'}
        }


def _group_sum_resolver(condition, key, operator, name, data_type, sub_conditions, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires a key: {condition}")

    if sub_conditions:
        return {
            name: {
                '$sum': {
                    '$cond': [
                        sub_conditions,
                        f'${key}',
                        0
                    ]
                }
            }
        }
    elif data_type == 'array':
        return {
            name: {'$sum': {'$sum': f'${key}'}}
        }
    else:
        return {
            name: {'$sum': f'${key}'}
        }


def _group_add_to_set_resolver(condition, key, operator, name, data_type, sub_conditions, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires a key: {condition}")

    if sub_conditions:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition's conditions not supported: {condition}")

    return {
        name: {'$addToSet': f'${key}'}
    }


def _group_merge_objects_resolver(condition, key, operator, name, data_type, sub_conditions, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires a key: {condition}")

    if sub_conditions:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition's conditions not supported: {condition}")

    return {
        name: {'$mergeObjects': f'${key}'}
    }


def _group_default_resolver(condition, key, operator, name, data_type, sub_conditions, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires a key: {condition}")

    if sub_conditions:
        raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition's conditions not supported: {condition}")

    if data_type == 'array':
        return {
            name: {f'${operator}': {f'${operator}': f'${key}'}}
        }
    else:
        return {
            name: {f'${operator}': f'${key}'}
        }


def _project_size_resolver(condition, key, operator, name, fields, group_keys, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.key' condition requires a key: {condition}")

    if key in group_keys:
        key = f'_id.{key}'

    return {
        name: {'$size': f'${key}'}
    }


def _project_sum_resolver(condition, key, operator, name, fields, group_keys, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.key' condition requires a key: {condition}")

    if key in group_keys:
        key = f'_id.{key}'

    return {
        name: {'$sum': f'${key}'}
    }


def _project_array_to_object_resolver(condition, key, operator, name, fields, group_keys, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition requires a key: {condition}")

    if key in group_keys:
        key = f'_id.{key}'

    return {
        name: {'$arrayToObject': f'${key}'}
    }


def _project_object_to_array_resolver(condition, key, operator, name, fields, group_keys, *args):
    if key is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition requires a key: {condition}")

    if key in group_keys:
        key = f'_id.{key}'

    return {
        name: {'$objectToArray': f'${key}'}
    }


def _project_calculate_sub_query(condition, operator, fields, group_keys, *args):
    supported_operator = ['add', 'subtract', 'multiply', 'divide']

    if operator is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.operator' condition requires a operator: {condition}")

    if operator not in supported_operator:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.operator' condition operator are not allowed"
                                    f" (supported_operator={supported_operator}): {condition}")

    if fields is None:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition requires a fields: {condition}")

    if operator in ['subtract', 'divide'] and len(fields) != 2:
        raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition requires two fields: {condition}")

    expressions = []
    for key in fields:
        if isinstance(key, str):
            if key in group_keys:
                key = f'_id.{key}'

            expressions.append(f'${key}')
        elif isinstance(key, (int, float)) and not isinstance(key, bool):
            expressions.append(key)
        elif isinstance(key, dict):
            sub_condition = key
            sub_operator = sub_condition.get('operator', sub_condition.get('o'))
            sub_fields = sub_condition.get('fields', sub_condition.get('f'))
            expressions.append(_project_calculate_sub_query(sub_condition, sub_operator, sub_fields, group_keys))
        else:
            raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition accept only string or object or "
                                        f"numeric types: {condition}")

    return {
        f'${operator}': expressions
    }


def _project_calculate_resolver(condition, key, operator, name, fields, group_keys, *args):
    return {
        name: _project_calculate_sub_query(condition, operator, fields, group_keys)
    }


STAT_GROUP_OPERATORS = {
    'count': _group_count_resolver,
    'sum': _group_sum_resolver,
    'average': _group_average_resolver,
    'max': _group_default_resolver,
    'min': _group_default_resolver,
    'first': _group_default_resolver,
    'last': _group_default_resolver,
    'push': _group_push_resolver,
    'add_to_set': _group_add_to_set_resolver,
    'merge_objects': _group_merge_objects_resolver,
}

STAT_PROJECT_OPERATORS = {
    'size': _project_size_resolver,
    'sum': _project_sum_resolver,
    'array_to_object': _project_array_to_object_resolver,
    'object_to_array': _project_object_to_array_resolver,
    'add': _project_calculate_resolver,
    'subtract': _project_calculate_resolver,
    'multiply': _project_calculate_resolver,
    'divide': _project_calculate_resolver,
}


import bson
import re
import logging
import certifi
from datetime import datetime, date
from dateutil.relativedelta import relativedelta
from functools import reduce
from mongoengine import EmbeddedDocumentField, EmbeddedDocument, Document, QuerySet, register_connection
from mongoengine.fields import DateField, DateTimeField, ComplexDateTimeField
from pymongo import ReadPreference
from mongoengine.errors import *
from spaceone.core import config
from spaceone.core import utils
from spaceone.core.error import *
from spaceone.core.model import BaseModel
from spaceone.core.model.mongo_model.filter_operator import FILTER_OPERATORS
from spaceone.core.model.mongo_model.stat_operator import STAT_GROUP_OPERATORS, STAT_PROJECT_OPERATORS

_REFERENCE_ERROR_FORMAT = r'Could not delete document \((\w+)\.\w+ refers to it\)'
_MONGO_CONNECTIONS = []
_MONGO_INIT_MODELS = []
_LOGGER = logging.getLogger(__name__)


def _raise_reference_error(class_name, message):
    m = re.findall(_REFERENCE_ERROR_FORMAT, message)
    if len(m) > 0:
        raise ERROR_EXIST_RESOURCE(parent=class_name, child=m[0])


class MongoCustomQuerySet(QuerySet):

    def last(self):
        return self.order_by('-id').first()

    def update(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], dict):
            kwargs.update(args[0])
        super().update(**kwargs)

    def increment(self, key, amount=1):
        key = key.replace('.', '__')
        inc_data = {
            f'inc__{key}': amount
        }

        super().update(**inc_data)

    def decrement(self, key, amount=1):
        key = key.replace('.', '__')
        dec_data = {
            f'dec__{key}': amount
        }

        super().update(**dec_data)

    def set_data(self, key, data):
        key = key.replace('.', '__')
        set_data = {
            f'set__{key}': data
        }

        super().update(**set_data)

    def unset_data(self, *keys):
        unset_data = {}

        for key in keys:
            key = key.replace('.', '__')
            unset_data[f'unset__{key}'] = 1

        super().update(**unset_data)

    def append(self, key, data):
        key = key.replace('.', '__')
        append_data = {
            f'push__{key}': data
        }

        super().update(**append_data)

    def remove(self, key, data):
        key = key.replace('.', '__')
        remove_data = {
            f'pull__{key}': data
        }
        super().update(**remove_data)


class MongoModel(Document, BaseModel):

    auto_create_index = True
    meta = {
        'abstract': True,
        'queryset_class': MongoCustomQuerySet,
        'auto_create_index': False
    }

    @classmethod
    def init(cls):
        global_conf = config.get_global()

        if not global_conf.get('MOCK_MODE', False):
            cls.connect()

            if cls not in _MONGO_INIT_MODELS:
                cls.auto_create_index = global_conf.get('DATABASE_AUTO_CREATE_INDEX', True)
                cls._create_index()

                _MONGO_INIT_MODELS.append(cls)

        cls.load_default_meta()

    @classmethod
    def load_default_meta(cls):
        cls._meta['datetime_fields'] = []
        for name, field in cls._fields.items():
            if isinstance(field, DateField):
                cls._meta['datetime_fields'].append(name)
            elif isinstance(field, DateTimeField):
                cls._meta['datetime_fields'].append(name)
            elif isinstance(field, ComplexDateTimeField):
                cls._meta['datetime_fields'].append(name)

    @classmethod
    def connect(cls):
        db_alias = cls._meta.get('db_alias', 'default')
        if db_alias not in _MONGO_CONNECTIONS:
            global_conf = config.get_global()
            databases = global_conf.get('DATABASES', {})

            if db_alias not in databases:
                raise ERROR_DB_CONFIGURATION(backend=db_alias)

            db_conf = databases[db_alias].copy()

            if 'read_preference' in db_conf:
                read_preference = getattr(ReadPreference, db_conf['read_preference'], None)
                if read_preference:
                    db_conf['read_preference'] = read_preference
                else:
                    del db_conf['read_preference']

            db_name: str = db_conf.get('db', '')
            db_name_prefix = global_conf.get('DATABASE_NAME_PREFIX', '')
            db_conf['db'] = f'{db_name_prefix}{db_name}'

            host: str = str(db_conf.get('host', '')).strip()
            if host.startswith('mongodb+srv://'):
                db_conf['tlsCAFile'] = certifi.where()

            register_connection(db_alias, **db_conf)

            _MONGO_CONNECTIONS.append(db_alias)

    @classmethod
    def _create_index(cls):
        if cls.auto_create_index:
            indexes = cls._meta.get('indexes', [])

            if len(indexes) > 0:
                _LOGGER.debug(f'Create MongoDB Indexes ({cls.__name__} Model: {len(indexes)} Indexes)')

                unique_fields = cls._get_unique_fields()

                for unique_field in unique_fields:
                    try:
                        cls.create_index({
                            'fields': unique_field,
                            'unique': True
                        })

                    except Exception as e:
                        _LOGGER.error(f'Unique Index Creation Failure: {e}')

                for index in indexes:
                    try:
                        cls.create_index(index)

                    except Exception as e:
                        if e.code != 85: # 85: IndexOptionsConflict
                            _LOGGER.error(f'Index Creation Failure: {e}')

    @classmethod
    def _get_unique_fields(cls):
        unique_fields = []
        for name, field in cls._fields.items():
            if field.unique:
                if isinstance(field.unique_with, str):
                    unique_fields.append([field.name, field.unique_with])
                elif isinstance(field.unique_with, list):
                    unique_fields.append([field.name] + field.unique_with)
                else:
                    unique_fields.append([field.name])

        return unique_fields

    @classmethod
    def create(cls, data):
        create_data = {}

        for name, field in cls._fields.items():
            if name in data:
                create_data[name] = data[name]
            else:
                generate_id = getattr(field, 'generate_id', None)
                if generate_id:
                    create_data[name] = utils.generate_id(generate_id)

                if getattr(field, 'auto_now', False):
                    create_data[name] = datetime.utcnow()
                elif getattr(field, 'auto_now_add', False):
                    create_data[name] = datetime.utcnow()

        for unique_field in cls._get_unique_fields():
            conditions = {}
            for f in unique_field:
                conditions[f] = data.get(f)
            vos = cls.filter(**conditions)
            if vos.count() > 0:
                raise ERROR_SAVE_UNIQUE_VALUES(keys=unique_field)

        try:
            new_vo = cls(**create_data).save()
        except Exception as e:
            raise ERROR_DB_QUERY(reason=e)

        return new_vo

    def update(self, data):
        updatable_fields = self._meta.get(
            'updatable_fields', list(
                filter(
                    lambda x: x != self._meta.get('id_field', 'id'), self._fields.keys()
                )
            )
        )

        for name, field in self._fields.items():
            if getattr(field, 'auto_now', False):
                if name not in data.keys():
                    data[name] = datetime.utcnow()

        for unique_field in self._get_unique_fields():
            conditions = {'pk__ne': self.pk}
            for f in unique_field:
                conditions[f] = data.get(f)

            vos = self.filter(**conditions)
            if vos.count() > 0:
                raise ERROR_SAVE_UNIQUE_VALUES(keys=unique_field)

        for key in list(data.keys()):
            if key not in updatable_fields:
                del data[key]

        if data != {}:
            try:
                super().update(**data)
                self.reload()
            except Exception as e:
                raise ERROR_DB_QUERY(reason=e)

        return self

    def delete(self):
        try:
            super().delete()
        except OperationError as e:
            _raise_reference_error(self.__class__.__name__, str(e))
            raise ERROR_DB_QUERY(reason=e)
        except Exception as e:
            raise ERROR_DB_QUERY(reason=e)

    def terminate(self):
        super().delete()

    def increment(self, key, amount=1):
        key = key.replace('.', '__')
        inc_data = {
            f'inc__{key}': amount
        }

        super().update(**inc_data)
        self.reload()
        return self

    def decrement(self, key, amount=1):
        key = key.replace('.', '__')
        dec_data = {
            f'dec__{key}': amount
        }

        super().update(**dec_data)
        self.reload()
        return self

    def set_data(self, key, data):
        key = key.replace('.', '__')
        set_data = {
            f'set__{key}': data
        }

        super().update(**set_data)
        self.reload()
        return self

    def unset_data(self, *keys):
        unset_data = {}

        for key in keys:
            key = key.replace('.', '__')
            unset_data[f'unset__{key}'] = 1

        super().update(**unset_data)
        self.reload()
        return self

    def append(self, key, data):
        key = key.replace('.', '__')
        append_data = {}

        field = getattr(self._fields.get(key, {}), 'field', None)
        if field and isinstance(field, EmbeddedDocumentField):
            reference_model = field.document_type_obj
            append_data[f'push__{key}'] = reference_model(**data)
        else:
            append_data[f'push__{key}'] = data

        super().update(**append_data)
        self.reload()
        return self

    def remove(self, key, data):
        key = key.replace('.', '__')
        remove_data = {
            f'pull__{key}': data
        }
        super().update(**remove_data)
        self.reload()
        return self

    @classmethod
    def get(cls, only=None, **conditions):
        vos = cls.filter(**conditions)

        if vos.count() == 0:
            keys = tuple(conditions.keys())
            values = tuple(conditions.values())

            if len(keys) == 1:
                raise ERROR_NOT_FOUND(key=keys[0], value=values[0])
            else:
                raise ERROR_NOT_FOUND(key=keys, value=values)

        if only:
            only = cls._remove_duplicate_only_keys(only)
            vos = vos.only(*only)

        return vos.first()

    @classmethod
    def filter(cls, **conditions):
        change_conditions = {}
        for key, value in conditions.items():
            if isinstance(value, list):
                change_conditions[f'{key}__in'] = value
            else:
                change_conditions[key] = value

        return cls.objects.filter(**change_conditions)

    def to_dict(self):
        return self.to_mongo()

    @classmethod
    def _get_target_objects(cls, target):
        if target:
            read_preference = getattr(ReadPreference, target, None)
            if read_preference:
                return cls.objects.read_preference(read_preference)

        return cls.objects

    @staticmethod
    def _check_operator_value(is_multiple, operator, value, condition):
        if is_multiple:
            if not isinstance(value, list):
                raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        else:
            if isinstance(value, list):
                raise ERROR_OPERATOR_VALUE_TYPE(operator=operator, condition=condition)

    @classmethod
    def _check_reference_field(cls, key):
        ref_keys = cls._meta.get('reference_query_keys', {}).keys()
        if key in ref_keys:
            return False
        else:
            return True

    @classmethod
    def _get_reference_model(cls, key):
        for ref_key, ref_conf in cls._meta.get('reference_query_keys', {}).items():
            if key.startswith(ref_key) and key[len(ref_key)] == '.':
                if isinstance(ref_conf, dict):
                    ref_model = ref_conf.get('model')
                    foreign_key = ref_conf.get('foreign_key')
                else:
                    ref_model = ref_conf
                    foreign_key = None

                ref_query_key = key.replace(f'{ref_key}.', '')
                if ref_model == 'self':
                    ref_model = cls

                return ref_model, ref_key, ref_query_key, foreign_key

        return None, None, None, None

    @classmethod
    def _change_reference_condition(cls, key, value, operator):
        ref_model, ref_key, ref_query_key, foreign_key = cls._get_reference_model(key)
        if ref_model:
            if value is None:
                return ref_key, value, operator
            else:
                ref_vos, total_count = ref_model.query(
                    filter=[{'k': ref_query_key, 'v': value, 'o': operator}])

                if foreign_key:
                    ref_values = []
                    for ref_vo in ref_vos:
                        ref_value = getattr(ref_vo, foreign_key)
                        if ref_value:
                            ref_values.append(ref_value)
                else:
                    ref_values = list(ref_vos)
                return ref_key, ref_values, 'in'

        else:
            return key, value, operator

    @classmethod
    def _make_condition(cls, condition):
        key = condition.get('key', condition.get('k'))
        value = condition.get('value', condition.get('v'))
        operator = condition.get('operator', condition.get('o'))
        change_query_keys = cls._meta.get('change_query_keys', {})

        if operator not in FILTER_OPERATORS:
            raise ERROR_DB_QUERY(reason=f'Filter operator is not supported. (operator = '
                                        f'{FILTER_OPERATORS.keys()})')

        resolver, mongo_operator, is_multiple = FILTER_OPERATORS.get(operator)

        cls._check_operator_value(is_multiple, operator, value, condition)

        if key and operator:
            if key in change_query_keys:
                key = change_query_keys[key]

            if operator not in ['regex', 'regex_in']:
                if cls._check_reference_field(key):
                    key, value, operator = cls._change_reference_condition(key, value, operator)

                    resolver, mongo_operator, is_multiple = FILTER_OPERATORS[operator]

                key = key.replace('.', '__')

            return resolver(key, value, mongo_operator, is_multiple)
        else:
            raise ERROR_DB_QUERY(reason='Filter condition should have key, value and operator.')

    @classmethod
    def _make_filter(cls, filter, filter_or):
        _filter = None
        _filter_or = None

        if len(filter) > 0:
            _filter = reduce(lambda x, y: x & y, map(cls._make_condition, filter))

        if len(filter_or) > 0:
            _filter_or = reduce(lambda x, y: x | y, map(cls._make_condition, filter_or))

        if _filter and _filter_or:
            _filter = _filter & _filter_or
        else:
            _filter = _filter or _filter_or

        return _filter

    @classmethod
    def _remove_duplicate_only_keys(cls, only):
        changed_only = []
        duplicated_only = []
        for key in only:
            exists = False

            for changed_key in changed_only:
                if key == changed_key or key.startswith(f'{changed_key}.'):
                    exists = True
                elif changed_key.startswith(f'{key}.'):
                    duplicated_only.append(changed_key)

            if exists is False:
                changed_only.append(key)

        if len(duplicated_only) > 0:
            changed_only = list(set(changed_only) - set(duplicated_only))

        return changed_only

    @classmethod
    def query(cls, *args, only=None, exclude=None, all_fields=False, filter=None, filter_or=None,
              sort=None, page=None, minimal=False, count_only=False, target=None, **kwargs):

        if filter is None:
            filter = []

        if filter_or is None:
            filter_or = []

        if sort is None:
            sort = {}

        if page is None:
            page = {}

        _order_by = []
        minimal_fields = cls._meta.get('minimal_fields')

        _filter = cls._make_filter(filter, filter_or)

        if 'key' in sort:
            if sort.get('desc', False):
                _order_by.append(f'-{sort["key"]}')
            else:
                _order_by.append(f'{sort["key"]}')
        elif 'keys' in sort:
            for s in sort['keys']:
                if s.get('desc', False):
                    _order_by.append(f'-{s["key"]}')
                else:
                    _order_by.append(f'{s["key"]}')

        try:
            vos = cls._get_target_objects(target).filter(_filter)

            if len(_order_by) > 0:
                vos = vos.order_by(*_order_by)

            if only:
                if len(_order_by) > 0:
                    ordering = _order_by
                else:
                    ordering = cls._meta.get('ordering')

                for key in ordering:
                    if key.startswith('+') or key.startswith('-'):
                        key = key[1:]
                    if key not in only:
                        only.append(key)

                only = cls._remove_duplicate_only_keys(only)
                vos = vos.only(*only)

            if exclude:
                vos = vos.exclude(*exclude)

            if minimal and minimal_fields:
                vos = vos.only(*minimal_fields)

            if all_fields:
                vos = vos.all_fields()

            total_count = vos.count()

            if count_only:
                vos = []

            else:
                if 'limit' in page and page['limit'] > 0:
                    start = page.get('start', 1)
                    if start < 1:
                        start = 1

                    vos = vos[start - 1:start + page['limit'] - 1]

            return vos, total_count

        except Exception as e:
            raise ERROR_DB_QUERY(reason=e)

    @classmethod
    def _check_well_known_type(cls, value):
        if isinstance(value, datetime):
            return f'{value.isoformat()}Z'
        elif isinstance(value, bson.objectid.ObjectId):
            return str(value)
        elif isinstance(value, Document):
            return str(value.id)
        elif isinstance(value, EmbeddedDocument):
            return dict(value.to_mongo())
        else:
            return value

    @classmethod
    def _make_aggregate_values(cls, cursor):
        values = []
        for row in cursor:
            data = {}
            for key, value in row.items():
                if key == '_id' and isinstance(row[key], dict):
                    for group_key, group_value in row[key].items():
                        data[group_key] = cls._check_well_known_type(group_value)
                else:
                    data[key] = cls._check_well_known_type(value)

            values.append(data)

        return values

    @classmethod
    def _make_distinct_values(cls, values):
        changed_values = []
        for value in values:
            changed_values.append(cls._check_well_known_type(value))

        return changed_values

    @classmethod
    def _make_sub_conditions(cls, sub_conditions, _before_group_keys):
        and_sub_conditions = []

        for sub_condition in sub_conditions:
            key = sub_condition.get('key', sub_condition.get('k'))
            value = sub_condition.get('value', sub_condition.get('v'))
            operator = sub_condition.get('operator', sub_condition.get('o'))

            if key is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.key' condition requires a key: {sub_condition}")

            if value is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.value' condition requires a value: {sub_condition}")

            if operator is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.operator' condition requires a operator: {sub_condition}")

            _SUPPORTED_OPERATOR = ['eq', 'not', 'gt', 'gte', 'lt', 'lte']

            if operator not in _SUPPORTED_OPERATOR:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.operator' condition's {operator} operator is not "
                           f"supported. (supported_operator = {_SUPPORTED_OPERATOR})")

            if key in _before_group_keys:
                key = f'_id.{key}'

            and_sub_conditions.append(
                {
                    f'${operator}': [f'${key}', value]
                }
            )

        return {
            '$and': and_sub_conditions
        }

    @classmethod
    def _get_group_fields(cls, condition, _before_group_keys):
        key = condition.get('key', condition.get('k'))
        name = condition.get('name', condition.get('n'))
        operator = condition.get('operator', condition.get('o'))
        sub_conditions = condition.get('conditions')
        sub_fields = condition.get('fields') or []
        data_type = condition.get('data_type')

        if operator not in STAT_GROUP_OPERATORS:
            raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition's {operator} operator is not supported. "
                                        f"(supported_operator = {list(STAT_GROUP_OPERATORS.keys())})")

        if name is None:
            raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields' condition requires a name: {condition}")

        if key in _before_group_keys:
            key = f'_id.{key}'

        for sub_field in sub_fields:
            f_key = sub_field.get('key', sub_field.get('k'))
            f_name = sub_field.get('name', sub_field.get('n'))

            if f_key is None:
                raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields.fields' condition requires a key: {condition}")

            if f_name is None:
                raise ERROR_DB_QUERY(reason=f"'aggregate.group.fields.fields' condition requires a name: {condition}")

            if f_key in _before_group_keys:
                sub_field['key'] = f'_id.{f_key}'

        if sub_conditions:
            sub_conditions = cls._make_sub_conditions(sub_conditions, _before_group_keys)

        return key, name, operator, sub_fields, sub_conditions, data_type

    @classmethod
    def _get_group_keys(cls, condition, _before_group_keys):
        key = condition.get('key', condition.get('k'))
        name = condition.get('name', condition.get('n'))
        date_format = condition.get('date_format')

        if key is None:
            raise ERROR_DB_QUERY(reason=f"'aggregate.group.keys' condition requires a key: {condition}")

        if name is None:
            raise ERROR_DB_QUERY(reason=f"'aggregate.group.keys' condition requires a name: {condition}")

        if key in _before_group_keys:
            key = f'_id.{key}'

        if date_format:
            if date_format == 'year':
                rule = {
                    '$year': f'${key}'
                }
            elif date_format == 'month':
                rule = {
                    '$month': f'${key}'
                }
            elif date_format == 'day':
                rule = {
                    '$dayOfMonth': f'${key}'
                }
            else:
                rule = {
                    '$dateToString': {
                        'format': date_format,
                        'date': f'${key}'
                    }
                }
        else:
            rule = f'${key}'

        return name, rule

    @classmethod
    def _make_group_rule(cls, options, _before_group_keys):
        _group_keys = []
        _group_rule = {
            '$group': {
                '_id': {}
            }
        }
        _keys = options.get('keys', [])
        _fields = options.get('fields', [])

        if len(_keys) == 0 and len(_fields) == 0:
            raise ERROR_REQUIRED_PARAMETER(key='aggregate.group.keys || aggregate.group.fields')

        for condition in _keys:
            name, rule = cls._get_group_keys(condition, _before_group_keys)
            _group_rule['$group']['_id'][name] = rule
            _group_keys.append(name)

        for condition in _fields:
            key, name, operator, sub_fields, sub_conditions, data_type = \
                cls._get_group_fields(condition, _before_group_keys)

            rule = STAT_GROUP_OPERATORS[operator](condition, key, operator, name, data_type, sub_conditions, sub_fields,
                                                  cls._meta['datetime_fields'])
            _group_rule['$group'].update(rule)

        return _group_rule, _group_keys

    @classmethod
    def _get_project_fields(cls, condition):
        key = condition.get('key', condition.get('k'))
        name = condition.get('name', condition.get('n'))
        operator = condition.get('operator', condition.get('o'))
        fields = condition.get('fields', condition.get('f'))

        if operator and operator not in STAT_PROJECT_OPERATORS:
            raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition's {operator} operator is not supported. "
                                        f"(supported_operator = {list(STAT_PROJECT_OPERATORS.keys())})")

        if name is None:
            raise ERROR_DB_QUERY(reason=f"'aggregate.project.fields' condition requires a name: {condition}")

        return key, name, operator, fields

    @classmethod
    def _make_project_rule(cls, options, _group_keys):
        _rules = []
        _fields = options.get('fields', [])
        _exclude_keys = options.get('exclude_keys', False)
        _project_rule = {
            '$project': {
            }
        }

        if len(_fields) == 0:
            raise ERROR_REQUIRED_PARAMETER(key='aggregate.project.fields')

        for condition in _fields:
            key, name, operator, fields = cls._get_project_fields(condition)

            if operator:
                rule = STAT_PROJECT_OPERATORS[operator](condition, key, operator, name, fields, _group_keys)
                _project_rule['$project'].update(rule)
            else:
                if key in _group_keys:
                    key = f'_id.{key}'

                _project_rule['$project'][name] = f'${key}'

        if _exclude_keys:
            _project_rule['$project']['_id'] = 0
            _group_keys = []

        return _project_rule, _group_keys

    @classmethod
    def _make_unwind_rule(cls, options):
        if 'path' not in options:
            raise ERROR_REQUIRED_PARAMETER(key='aggregate.unwind.path')

        return {
            '$unwind': f"${options['path']}"
        }

    @classmethod
    def _make_count_rule(cls, options):
        if 'name' not in options:
            raise ERROR_REQUIRED_PARAMETER(key='aggregate.count.name')

        return {
            '$count': options['name']
        }

    @classmethod
    def _make_sort_rule(cls, options, _group_keys):
        key = options.get('key')
        desc = options.get('desc', False)
        keys = options.get('keys')

        order_by = {}

        if key:
            if key in _group_keys:
                sort_key = f'_id.{key}'
            else:
                sort_key = key

            order_by[sort_key] = -1 if desc else 1

        elif keys:
            for k in keys:
                if k['key'] in _group_keys:
                    sort_key = f'_id.{k["key"]}'
                else:
                    sort_key = k['key']

                order_by[sort_key] = -1 if k.get('desc', False) else 1

        else:
            raise ERROR_REQUIRED_PARAMETER(key='aggregate.sort.key')

        return {
            '$sort': order_by
        }

    @classmethod
    def _make_limit_rule(cls, options):
        return {
            '$limit': options
        }

    @classmethod
    def _make_skip_rule(cls, options):
        return {
            '$skip': options
        }

    @classmethod
    def _make_aggregate_rules(cls, aggregate):
        _aggregate_rules = []
        _group_keys = []

        if not isinstance(aggregate, list):
            raise ERROR_INVALID_PARAMETER_TYPE(key='aggregate', type='list')

        for stage in aggregate:
            if 'unwind' in stage:
                rule = cls._make_unwind_rule(stage['unwind'])
                _aggregate_rules.append(rule)
            elif 'group' in stage:
                rule, group_keys = cls._make_group_rule(stage['group'], _group_keys)
                _aggregate_rules.append(rule)
                _group_keys += group_keys
            elif 'count' in stage:
                rule = cls._make_count_rule(stage['count'])
                _aggregate_rules.append(rule)
            elif 'sort' in stage:
                rule = cls._make_sort_rule(stage['sort'], _group_keys)
                _aggregate_rules.append(rule)
            elif 'project' in stage:
                rule, _group_keys = cls._make_project_rule(stage['project'], _group_keys)
                _aggregate_rules.append(rule)
            elif 'limit' in stage:
                rule = cls._make_limit_rule(stage['limit'])
                _aggregate_rules.append(rule)
            elif 'skip' in stage:
                rule = cls._make_skip_rule(stage['skip'])
                _aggregate_rules.append(rule)
            else:
                raise ERROR_REQUIRED_PARAMETER(key='aggregate.unwind or aggregate.group or '
                                                   'aggregate.count or aggregate.sort or '
                                                   'aggregate.project or aggregate.limit or '
                                                   'aggregate.skip')

        return _aggregate_rules

    @classmethod
    def _stat_aggregate(cls, vos, aggregate, page, allow_disk_use):
        result = {}
        pipeline = []
        _aggregate_rules = cls._make_aggregate_rules(aggregate)

        for rule in _aggregate_rules:
            pipeline.append(rule)

        if 'limit' in page and page['limit'] > 0:
            limit = page['limit']
            start = page.get('start', 1)
            start = 1 if start < 1 else start

            result['total_count'] = 0
            cursor = vos.aggregate(pipeline + [{'$count': 'total_count'}])
            for c in cursor:
                result['total_count'] = c['total_count']
                break

            if start > 1:
                pipeline.append({
                    '$skip': start - 1
                })

            pipeline.append({
                '$limit': limit
            })

        if allow_disk_use:
            _LOGGER.debug(f'[_stat_aggregate] allow_disk_use: {allow_disk_use}')
            cursor = vos.aggregate(pipeline, allowDiskUse=True)
        else:
            cursor = vos.aggregate(pipeline)
        result['results'] = cls._make_aggregate_values(cursor)
        return result

    @classmethod
    def _stat_distinct(cls, vos, distinct, page):
        result = {}
        values = vos.distinct(distinct)

        try:
            values.sort()
        except Exception:
            pass

        if 'limit' in page and page['limit'] > 0:
            start = page.get('start', 1)
            if start < 1:
                start = 1

            result['total_count'] = len(values)
            values = values[start - 1:start + page['limit'] - 1]

        result['results'] = cls._make_distinct_values(values)
        return result

    @classmethod
    def stat(cls, *args, aggregate=None, distinct=None, filter=None, filter_or=None, page=None,
             target='SECONDARY_PREFERRED', allow_disk_use=False, **kwargs):

        if filter is None:
            filter = []

        if filter_or is None:
            filter_or = []

        if page is None:
            page = {}

        if not (aggregate or distinct):
            raise ERROR_REQUIRED_PARAMETER(key='aggregate')

        _filter = cls._make_filter(filter, filter_or)

        try:
            vos = cls._get_target_objects(target).filter(_filter)

            if aggregate:
                return cls._stat_aggregate(vos, aggregate, page, allow_disk_use)

            elif distinct:
                return cls._stat_distinct(vos, distinct, page)

        except Exception as e:
            if not isinstance(e, ERROR_BASE):
                e = ERROR_UNKNOWN(message=str(e))

            raise ERROR_DB_QUERY(reason=e.message)

    @classmethod
    def _check_field_group(cls, field_group):
        for key in field_group:
            if key.startswith('_total_'):
                raise ERROR_INVALID_PARAMETER(key='field_group',
                                              reason='Field group keys cannot contain _total_ characters.')

    @classmethod
    def _make_group_keys(cls, group_by, date_field, granularity=None):
        group_keys = []
        for key in group_by:
            name = key.rsplit('.', 1)[-1:][0]
            group_keys.append({
                'key': key,
                'name': name
            })

        if granularity and granularity in ['DAILY', 'MONTHLY', 'YEARLY']:
            group_keys.append({
                'key': date_field,
                'name': 'date'
            })

        return group_keys

    @classmethod
    def _make_group_fields(cls, fields):
        group_fields = []
        for name, condition in fields.items():
            cls._check_condition(condition)
            operator = condition['operator']
            key = condition.get('key')
            fields = condition.get('fields')
            data_type = condition.get('data_type')

            group_field = {
                'name': name,
                'operator': operator,
                'data_type': data_type
            }

            if operator != 'count':
                group_field['key'] = key

            if operator == 'push' and fields:
                push_fields = []
                for key, value in fields.items():
                    push_fields.append({
                        'key': value,
                        'name': key
                    })

                group_field['fields'] = push_fields

            group_fields.append(group_field)

        return group_fields

    @classmethod
    def _check_condition(cls, condition):
        key = condition.get('key')
        operator = condition.get('operator')
        fields = condition.get('fields', [])

        if operator is None:
            raise ERROR_REQUIRED_PARAMETER(key='query.fields.operator')

        # Check Operator
        if operator not in ['count', 'push'] and key is None:
            raise ERROR_REQUIRED_PARAMETER(key='query.fields.key')

        if operator == 'push' and not key and len(fields) == 0:
            raise ERROR_REQUIRED_PARAMETER(key='query.fields.fields')

    @classmethod
    def _make_field_group_keys(cls, group_keys, field_group):
        field_group_keys = []
        for group_key in group_keys:
            key = group_key['key'].rsplit('.', 1)[-1:][0]
            name = group_key['name']
            if name not in field_group:
                if name == 'date':
                    field_group_keys.append({
                        'key': 'date',
                        'name': 'date'
                    })
                else:
                    field_group_keys.append({
                        'key': key,
                        'name': name
                    })

        return field_group_keys

    @classmethod
    def _make_field_group_fields(cls, group_fields, field_group):
        field_group_fields = []
        for group_field in group_fields:
            field_name = group_field['name']
            operator = group_field['operator']

            if operator in ['sum', 'average', 'max', 'min', 'count']:
                field_group_fields.append({
                    'key': field_name,
                    'name': f'_total_{field_name}',
                    'operator': 'sum' if operator == 'count' else operator
                })

            field_group_fields.append({
                'name': field_name,
                'operator': 'push',
                'fields': cls._make_field_group_push_fields(field_name, field_group)
            })

        return field_group_fields

    @classmethod
    def _make_field_group_push_fields(cls, field_name, field_group):
        push_fields = [
            {
                'name': 'value',
                'key': field_name
            }
        ]

        for field_group_key in field_group:
            push_fields.append({
                'name': field_group_key,
                'key': field_group_key
            })

        return push_fields

    @classmethod
    def _make_field_group_query(cls, group_keys, group_fields, field_group):
        field_group_query = {
            'group': {
                'keys': cls._make_field_group_keys(group_keys, field_group),
                'fields': cls._make_field_group_fields(group_fields, field_group)
            }
        }

        return [field_group_query]

    @classmethod
    def _make_select_query(cls, select):
        select_query = {
            'project': {
                'fields': [],
                'exclude_keys': True
            }
        }

        supported_operator = ['add', 'subtract', 'multiply', 'divide', 'size', 'sum']

        for name, condition in select.items():
            if isinstance(condition, str):
                select_query['project']['fields'].append({
                    'name': name,
                    'key': condition
                })
            elif isinstance(condition, dict):
                operator = condition.get('operator')
                key = condition.get('key')
                fields = condition.get('fields')

                if operator not in supported_operator:
                    raise ERROR_INVALID_PARAMETER(key='query.select.operator',
                                                  reason=f'supported_operator = {supported_operator}')

                if key:
                    select_query['project']['fields'].append({
                        'name': name,
                        'operator': operator,
                        'key': key
                    })
                elif fields:
                    select_query['project']['fields'].append({
                        'name': name,
                        'operator': operator,
                        'fields': fields
                    })
                else:
                    raise ERROR_REQUIRED_PARAMETER(key='query.select.fields')

        return [select_query]

    @classmethod
    def _make_sort_query(cls, sort, group_fields, has_field_group):
        sort_query = {
            'sort': {
                'keys': []
            }
        }
        if has_field_group:
            group_field_names = []
            for group_field in group_fields:
                group_field_names.append(group_field['name'])

            for condition in sort:
                key = condition.get('key')
                desc = condition.get('desc', False)

                if key in group_field_names:
                    key = f'_total_{key}'

                sort_query['sort']['keys'].append({
                    'key': key,
                    'desc': desc
                })
        else:
            sort_query['sort']['keys'] = sort

        return [sort_query]

    @classmethod
    def _make_page_query(cls, page):
        page_query = []
        start = page.get('start')
        limit = page.get('limit')

        if limit:
            if start:
                page_query.append({
                    'skip': start - 1
                })

            page_query.append({
                'limit': limit + 1
            })

        return page_query

    @classmethod
    def _make_date_filter(cls, date_field, key, operator):
        return [{
            'k': date_field,
            'v': key,
            'o': operator
        }]

    @classmethod
    def _parse_start_and_end_time(cls, key, value) -> date:
        if isinstance(value, date):
            return value
        elif isinstance(value, datetime):
            return value.date()
        elif isinstance(value, str):

            if len(value) == 4:
                date_format = '%Y'
                date_type = 'YYYY'
            elif len(value) == 7:
                date_format = '%Y-%m'
                date_type = 'YYYY-MM'
            else:
                date_format = '%Y-%m-%d'
                date_type = 'YYYY-MM-DD'

            try:
                dt = datetime.strptime(value, date_format).date()

                if date_type == 'YYYY':
                    if key == 'start':
                        return datetime(dt.year, 1, 1)
                    else:
                        return datetime(dt.year, 1, 1) + relativedelta(years=1)
                elif date_type == 'YYYY-MM':
                    if key == 'start':
                        return datetime(dt.year, dt.month, 1)
                    else:
                        return datetime(dt.year, dt.month, 1) + relativedelta(months=1)
                else:
                    if key == 'start':
                        return datetime(dt.year, dt.month, dt.day)
                    else:
                        return datetime(dt.year, dt.month, dt.day) + relativedelta(days=1)

            except Exception as e:
                raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_type)
        else:
            raise ERROR_INVALID_PARAMETER(key=key, reason=f'{key} option should be datetime or str')

    @classmethod
    def _convert_date_value(cls, date_value, date_field_format):
        if isinstance(date_field_format, str):
            return date_value.strftime(date_field_format)
        else:
            return date_value

    @classmethod
    def analyze(cls, *args, granularity=None, fields=None, select=None, group_by=None, field_group=None,
                filter=None, filter_or=None, page=None, sort=None, start=None, end=None,
                date_field='date', date_field_format='%Y-%m-%d', target='SECONDARY_PREFERRED',
                allow_disk_use=False, **kwargs):

        if fields is None:
            raise ERROR_REQUIRED_PARAMETER(key='fields')

        filter = filter or []
        filter_or = filter_or or []
        group_by = group_by or []
        sort = sort or []
        page = page or {}
        field_group = field_group or []
        cls._check_field_group(field_group)
        has_field_group = len(field_group) > 0
        page_limit = page.get('limit')

        if start:
            start_time: date = cls._parse_start_and_end_time('start', start)
            start_value = cls._convert_date_value(start_time, date_field_format)
            filter += cls._make_date_filter(date_field, start_value, 'gte')

        if end:
            end_time: date = cls._parse_start_and_end_time('end', end)
            end_value = cls._convert_date_value(end_time, date_field_format)
            filter += cls._make_date_filter(date_field, end_value, 'lt')

        group_keys = cls._make_group_keys(group_by, date_field, granularity)
        group_fields = cls._make_group_fields(fields)

        query = {
            'filter': filter,
            'filter_or': filter_or,
            'aggregate': [
                {
                    'group': {
                        'keys': group_keys,
                        'fields': group_fields
                    }
                }
            ],
            'target': target,
            'allow_disk_use': allow_disk_use
        }

        if select:
            query['aggregate'] += cls._make_select_query(select)

        if has_field_group:
            query['aggregate'] += cls._make_field_group_query(group_keys, group_fields, field_group)

        if len(sort) > 0:
            query['aggregate'] += cls._make_sort_query(sort, group_fields, has_field_group)

        if page:
            query['aggregate'] += cls._make_page_query(page)

        response = cls.stat(**query)

        if page_limit:
            response['more'] = len(response['results']) > page_limit
            response['results'] = response['results'][:page_limit]

        return response



from spaceone.core.pygrpc.client import client, get_grpc_method
from spaceone.core.pygrpc.server import serve
from spaceone.core.pygrpc.api import BaseAPI

__all__ = ['serve', 'client', 'get_grpc_method', 'BaseAPI']



from spaceone.api.sample.v1 import helloworld_pb2

__all__ = ['HelloWorldInfo']


def HelloWorldInfo(message):
    info = {
        'message': message
    }

    return helloworld_pb2.HelloReply(**info)



import json
import logging

from spaceone.core.cache import cacheable
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core.auth.jwt import JWTAuthenticator, JWTUtil
from spaceone.core.transaction import get_transaction
from spaceone.core.handler import BaseAuthenticationHandler
from spaceone.core.error import ERROR_AUTHENTICATE_FAILURE, ERROR_HANDLER_CONFIGURATION

_LOGGER = logging.getLogger(__name__)


class AuthenticationGRPCHandler(BaseAuthenticationHandler):

    def __init__(self, config):
        super().__init__(config)
        self.client = None
        self._initialize()

    def _initialize(self):
        if 'uri' not in self.config:
            _LOGGER.error(f'[_initialize] uri config is undefined.')
            raise ERROR_HANDLER_CONFIGURATION(handler='AuthenticationGRPCHandler')

        self.client: SpaceConnector = self.locator.get_connector(SpaceConnector, endpoint=self.config['uri'])

    def verify(self, params=None):
        token = self._get_token()
        domain_id = self._extract_domain_id(token)

        token_info = self._authenticate(token, domain_id)
        self._update_meta(token_info)

    @cacheable(key='public-key:{domain_id}', backend='local')
    def _get_public_key(self, domain_id):
        _LOGGER.debug(f'[_get_public_key] get jwk from identity service ({domain_id})')
        response = self.client.dispatch('Domain.get_public_key', {'domain_id': domain_id})

        return response['public_key']

    def _authenticate(self, token, domain_id):
        public_key = self._get_public_key(domain_id)

        payload = JWTAuthenticator(json.loads(public_key)).validate(token)
        # TODO: if payload is api_key type and record is delete in database, raise ERROR_AUTHENTICATE_FAILURE exception.

        return payload

    @staticmethod
    def _get_token():
        transaction = get_transaction()
        token = transaction.meta.get('token')
        if not isinstance(token, str) \
                or token is None \
                or len(token) == 0:
            raise ERROR_AUTHENTICATE_FAILURE(message='Empty token provided.')
        return token

    @staticmethod
    def _extract_domain_id(token):
        try:
            decoded = JWTUtil.unverified_decode(token)
        except Exception:
            _LOGGER.debug(f'[_extract_domain_id] failed to decode token: {token[:10]}')
            raise ERROR_AUTHENTICATE_FAILURE(message='Cannot decode token.')

        domain_id = decoded.get('did')

        if domain_id is None:
            raise ERROR_AUTHENTICATE_FAILURE(message='Empty domain_id provided.')

        return domain_id

    def _update_meta(self, token_info):
        domain_id = token_info.get('did')
        user_id = token_info.get('aud')
        user_type = token_info.get('user_type')
        token_type = token_info.get('cat')
        api_key_id = token_info.get('api_key_id')
        permissions = token_info.get('permissions')

        self.transaction.set_meta('domain_id', domain_id)
        self.transaction.set_meta('user_id', user_id)
        self.transaction.set_meta('authorization.user_type', user_type)
        self.transaction.set_meta('authorization.permissions', permissions)
        self.transaction.set_meta('token_type', token_type)
        self.transaction.set_meta('api_key_id', api_key_id)



# -*- coding: utf-8 -*-
import logging
import json


class MessageJsonFilter(logging.Filter):
    def filter(self, record):
        if getattr(record, 'msg', None):
            record.msg_dump = json.dumps(str(record.msg))
        else:
            record.msg_dump = "\"\""

        return True



from cachetools import TTLCache

from spaceone.core.error import *
from spaceone.core.cache import BaseCache


class LocalCache(BaseCache):

    def __init__(self, backend, cache_conf):
        try:
            max_size = cache_conf.get('max_size', 128)
            ttl = cache_conf.get('ttl', 86400)
            self.cache = TTLCache(maxsize=max_size, ttl=ttl)
        except Exception:
            raise ERROR_CACHE_CONFIGURATION(backend=backend)

    def get(self, key, **kwargs):
        return self.cache.get(key)

    def set(self, key, value, expire=None, **kwargs):
        if expire:
            raise ERROR_CACHE_OPTION(method='cache.set', option='expire')

        self.cache[key] = value
        return True

    def delete(self, *keys):
        for key in keys:
            del self.cache[key]

    def flush(self, is_async=False):
        self.cache.popitem()
        self.cache.expire()



import logging

from spaceone.core import config
from spaceone.core.error import *

__init__ = ['put', 'get']

_QUEUE_CONNECTIONS = {}
LOGGER = logging.getLogger(__name__)


def _create_connection(topic):
    global_conf = config.get_global()
    if 'QUEUES' not in global_conf:
        raise ERROR_CONFIGURATION(key='QUEUES')

    queue_conf = global_conf['QUEUES'][topic].copy()
    backend = queue_conf.pop('backend', None)
    module_name, class_name = backend.rsplit('.', 1)
    queue_module = __import__(module_name, fromlist=[class_name])
    return getattr(queue_module, class_name)(queue_conf)


def connection(func):
    def wrapper(*args, **kwargs):
        topic = args[0]
        new_args = tuple(list(args[1:]))
        if topic not in _QUEUE_CONNECTIONS:
            _QUEUE_CONNECTIONS[topic] = _create_connection(topic)

        return func(_QUEUE_CONNECTIONS[topic], *new_args, **kwargs)

    return wrapper


@connection
def get(queue_cls):
    return queue_cls.get()


@connection
def put(queue_cls, key):
    return queue_cls.put(key)


class BaseQueue(object):
    initialized = None

    def initialize(self):
        """
        Re initilize if there are problem
        """
        pass

    def get(self):
        """
        Returns:
            queue_value(any)
        """
        raise NotImplementedError('queue.get not implemented!')

    def put(self, key):
        """
        Args:
            key(any)

        Returns:
            True | False
        """
        raise NotImplementedError('queue.set not implemented!')



from spaceone.core.base import CoreObject


class BaseManager(CoreObject):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        for key, value in kwargs.items():
            setattr(self, key, value)



import logging
import logging.config
import copy

from spaceone.core import config
from spaceone.core.error import *
from spaceone.core.utils import *
from spaceone.core.logger.filters import *

__all__ = ['set_logger']

DEFAULT_LOGGER = 'spaceone'

HANDLER_DEFAULT_CONSOLE = {
    'class': 'logging.StreamHandler',
    'formatter': 'standard',
    'filters': ['transaction', 'masking', 'exclude', 'parameter', 'traceback']
}

HANDLER_DEFAULT_FILE = {
    'class': 'logging.handlers.RotatingFileHandler',
    'filename': '',
    'filters': ['transaction', 'masking', 'exclude', 'error_message', 'parameter_log', 'message', 'traceback_log'],
    'formatter': 'file',
    'maxBytes': 10485760,  # 10 MB
    'backupCount': 10
}

HANDLER_DEFAULT_TMPL = {
    'console': HANDLER_DEFAULT_CONSOLE,
    'file': HANDLER_DEFAULT_FILE
}

FORMATTER_DEFAULT_TMPL = {
    'standard': {
        'format': '%(asctime)s.%(msecs)03dZ [%(levelname)s] %(trace_id)s %(domain_id)s %(user_id)s %(tnx_method)s (%(filename)s:%(lineno)d) %(message)s %(params)s %(traceback)s',
        'datefmt': '%Y-%m-%dT%H:%M:%S',
    },
    'file': {
        'format': '{"time": "%(asctime)s.%(msecs)03dZ", "level": "%(levelname)s", "peer": "%(peer)s", "trace_id": "%(trace_id)s", "domain_id": "%(domain_id)s", "user_id": "%(user_id)s", "tnx_method": "%(tnx_method)s", "file_name": "%(filename)s", "line": %(lineno)d, "parameter": %(params_log)s, "message": %(msg_dump)s, "error": { "code": "%(error_code)s", "message": "%(error_message)s", "traceback": %(traceback_log)s }}',
        'datefmt': '%Y-%m-%dT%H:%M:%S',
    }
}

FILTER_DEFAULT_TMPL = {
    'masking': {
        '()': MaskingFilter,
        'rules': {}
    },
    'transaction': {
        '()': TransactionFilter
    },
    'traceback': {
        '()': TracebackFilter
    },
    'traceback_log': {
        '()': TracebackLogFilter
    },
    'parameter': {
        '()': ParameterFilter
    },
    'parameter_log': {
        '()': ParameterLogFilter
    },
    'error_message': {
        '()': ErrorFilter
    },
    'message': {
        '()': MessageJsonFilter
    },
    'exclude': {
        '()': ExcludeFilter,
        'rules': {
            'tnx_method': [],
        }
    }
}

_LOGGER = {
    'version': 1,
    'formatters': {},
    'filters': {},
    'handlers': {
        'console': HANDLER_DEFAULT_CONSOLE
    },
    'loggers': {}
}

LOGGER_DEFAULT_TMPL = {
    'level': 'DEBUG',
    'propagate': True,
    'handlers': ['console']
}


def set_logger(transaction=None):
    _set_config(transaction)
    logging.config.dictConfig(_LOGGER)


def _set_default_logger(default_logger):
    _LOGGER['loggers'] = {default_logger: LOGGER_DEFAULT_TMPL}
    _LOGGER['formatters'] = FORMATTER_DEFAULT_TMPL


def _set_loggers(loggers):
    for _logger in loggers:
        _LOGGER['loggers'][_logger] = deep_merge(loggers[_logger], copy.deepcopy(LOGGER_DEFAULT_TMPL))


def _set_transaction_filter(transaction):
    if transaction:
        _LOGGER['filters']['transaction']['transaction'] = transaction


def _set_handlers(handlers):
    for _handler in handlers:
        _default = copy.deepcopy(HANDLER_DEFAULT_TMPL)

        if 'type' in handlers[_handler]:
            if handlers[_handler]['type'] not in HANDLER_DEFAULT_TMPL:
                raise ERROR_LOG_CONFIG(reason='Logger handler type is not supported')

            _default = copy.deepcopy(HANDLER_DEFAULT_TMPL[handlers[_handler]['type']])

        _default = deep_merge(handlers[_handler], _default)

        if 'type' in _default:
            del _default['type']

        _LOGGER['handlers'][_handler] = _default


def _set_formatters(formatters):
    for _formatter in formatters:
        _default = {}

        if 'type' in formatters[_formatter]:
            if formatters[_formatter]['type'] not in FORMATTER_DEFAULT_TMPL:
                raise ERROR_LOG_CONFIG(reason='Logger formatter type is not supported')

            _default = copy.deepcopy(FORMATTER_DEFAULT_TMPL[formatters[_formatter]['type']])

        _default = deep_merge(formatters[_formatter]['args'], _default)

        if 'type' in _default:
            del _default['type']

        _LOGGER['formatters'][_formatter] = _default


def _set_filters(filters):
    _LOGGER['filters'] = deep_merge(filters, copy.deepcopy(FILTER_DEFAULT_TMPL))


def _set_config(transaction):
    global_log_conf = config.get_global('LOG', {})

    _set_default_logger(DEFAULT_LOGGER)

    if 'loggers' in global_log_conf:
        _set_loggers(global_log_conf['loggers'])

    if 'handlers' in global_log_conf:
        _set_handlers(global_log_conf['handlers'])

    if 'formatters' in global_log_conf:
        _set_formatters(global_log_conf['formatters'])

    _set_filters(global_log_conf.get('filters', {}))
    _set_transaction_filter(transaction)



import logging

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from spaceone.core import config, utils
from spaceone.core.logger import set_logger
from spaceone.core.opentelemetry import set_tracer, set_metric
from spaceone.core.extension.server_info import ServerInfoManager

_LOGGER = logging.getLogger(__name__)


def _get_router_conf():
    package = config.get_package()
    router_conf_module = __import__(f'{package}.conf.router_conf', fromlist=['router_conf'])
    return getattr(router_conf_module, 'ROUTER', [])


def _get_sub_app_conf():
    package = config.get_package()
    router_conf_module = __import__(f'{package}.conf.router_conf', fromlist=['router_conf'])
    return getattr(router_conf_module, 'SUB_APP', {})


def _get_router(path):
    try:
        module_path, router_name = path.split(':')
        module_name = module_path.rsplit('.')[-1:]
        router_module = __import__(module_path, fromlist=[module_name])
        return getattr(router_module, router_name)
    except Exception as e:
        _LOGGER.warning(f'[_get_router] Invalid router path. (router_path = {path})', exc_info=True)


def _mount_sub_apps(app, sub_apps):
    sub_app_conf = _get_sub_app_conf()
    for sub_app_name, sub_app_options in sub_app_conf.items():
        sub_app = sub_apps.get(sub_app_name)
        if sub_app:
            sub_app_path = sub_app_options.get('path')
            app.mount(path=sub_app_path, app=sub_app)


def _create_sub_app(sub_app_options):
    title = sub_app_options.get('title', 'FastAPI')
    description = sub_app_options.get('description', '')
    contact = sub_app_options.get('contact', {})

    return FastAPI(title=title, description=description, contact=contact)


def _include_routers(app):
    # Include All Routers from router_conf.py
    routers_conf = _get_router_conf()
    sub_apps_conf = _get_sub_app_conf()

    sub_apps = {}
    all_routers_path = []

    # App Routers
    for router_conf in routers_conf:
        sub_app_name = router_conf.get('sub_app')
        router_path = router_conf.get('router_path')
        router_options = router_conf.get('router_options', {})

        if router_path is None:
            _LOGGER.warning(f'[include_routers] Undefined router_path. (router = {router_conf})')
            continue

        if sub_app_name in sub_apps_conf:
            if sub_app_name not in sub_apps:
                sub_app_options = sub_apps_conf[sub_app_name]
                sub_apps[sub_app_name] = _create_sub_app(sub_app_options)

            _append_router(sub_apps[sub_app_name], router_path, router_options)
        else:
            _append_router(app, router_path, router_options)

        all_routers_path.append(router_path)

    # Extension Routers
    ext_routers = config.get_global('REST_EXTENSION_ROUTERS', [])
    for router in ext_routers:
        router_path = router.get('router_path')
        router_options = router.get('router_options', {})
        _append_router(app, router_path, router_options)

        all_routers_path.append(router_path)

    # Mount Sub Applications
    _mount_sub_apps(app, sub_apps)

    all_routers_path_str = '\n\t - '.join(all_routers_path)
    _LOGGER.debug(f'Loaded Routers: \n\t - {all_routers_path_str}')

    return app


def _append_router(app, router_path, router_options):
    router = _get_router(router_path)

    app.include_router(
        router,
        **router_options
    )


def _add_middlewares(app):
    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )
    return app


def _init_fast_api():
    global_conf = config.get_global()
    server_info = ServerInfoManager()

    return FastAPI(
        title=global_conf.get('REST_TITLE', 'Document'),
        version=server_info.get_version(),
        contact=global_conf.get('REST_CONTACT', {}),
        description=global_conf.get('REST_DESCRIPTION', ''),
    )


def fast_api_app():
    app = _init_fast_api()
    app = _add_middlewares(app)
    app = _include_routers(app)
    return app


def serve():
    conf = config.get_global()

    # Enable logging configuration\
    set_logger()

    # Set OTel Tracer and Metric|
    set_tracer()

    uvicorn_options = conf.get('UVICORN_OPTIONS', {})

    _LOGGER.info(f'Start REST Server ({config.get_service()}): '
                 f'host={conf["HOST"]} port={conf["PORT"]} options={uvicorn_options}')

    uvicorn.run('spaceone.core.fastapi.server:fast_api_app', host=conf['HOST'], port=conf['PORT'], **uvicorn_options)



import logging
import os
import pkg_resources

from spaceone.core import config
from spaceone.api.core.v1 import server_info_pb2, server_info_pb2_grpc

_LOGGER = logging.getLogger(__name__)


class ServerInfo(server_info_pb2_grpc.ServerInfoServicer):

    def __init__(self):
        self.server_info = ServerInfoManager()

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def pb2_grpc_module(self):
        return server_info_pb2_grpc

    @property
    def service_name(self):
        return server_info_pb2.DESCRIPTOR.services_by_name[self.__class__.__name__].full_name

    def get_version(self, request, context):
        return server_info_pb2.VersionInfo(version=self.server_info.get_version())


class ServerInfoManager(object):

    def get_version(self):
        service = config.get_service()
        version = self._get_version_from_pkg(service)
        if version is None:
            version = self._get_version_from_file(service)

        return version or 'UNKNOWN'

    @staticmethod
    def _get_version_from_pkg(service):
        try:
            # TODO: Change the process of finding the package name
            return pkg_resources.get_distribution(f'spaceone-{service}').version
        except Exception as e:
            return None

    @staticmethod
    def _get_version_from_file(service):
        try:
            module = __import__(f'spaceone.{service}', fromlist=[service])
            root_path = os.path.dirname(os.path.dirname(os.path.dirname(module.__file__)))
            with open(f'{root_path}/VERSION', 'r') as f:
                version = f.read().strip()
                f.close()
                return version
        except Exception as e:
            return None



import logging
from opentelemetry import trace
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from spaceone.core import config

__all__ = ['set_tracer']

_LOGGER = logging.getLogger(__name__)


def set_tracer():
    service = config.get_service()
    endpoint = config.get_global('OTEL', {}).get('endpoint')

    resource = Resource(attributes={
        SERVICE_NAME: service
    })
    provider = TracerProvider(resource=resource)

    if endpoint:
        processor = BatchSpanProcessor(OTLPSpanExporter(endpoint=endpoint))
    else:
        processor = BatchSpanProcessor(SpanExporter())

    provider.add_span_processor(processor)
    trace.set_tracer_provider(provider)



import redis
import time
import logging

from spaceone.core.error import *
from spaceone.core.queue import BaseQueue

_LOGGER = logging.getLogger(__name__)

# Wait max 1 minute to recover
MAX_TRY = 6
WAIT_INTERVAL = 10


class RedisQueue(BaseQueue):
    """
    This is multiprocessing Queue for python multiprocessing
    Does not support Queue between different processes
    """
    def __init__(self, conf):
        self._conf = conf
        try:
            self.channel = conf.pop('channel')
            pool = redis.ConnectionPool(**conf)
            self.conn = redis.Redis(connection_pool=pool)
            self.conn.ping()

            self.pubsub = self.conn.pubsub()
            self.pubsub.subscribe(self.channel)
            self.initialized = True
        except Exception:
            for i in range(MAX_TRY):
                _LOGGER.error(f"######### queue connection check : {i} ###########")
                self.initialize()
                if self.initialized:
                    return
                time.sleep(WAIT_INTERVAL)
            self.initialized = False
            _LOGGER.error(f"run without configuration for late recover: {conf}")
            #raise ERROR_CONFIGURATION(key="backend")

    def initialize(self):
        """ Sometime initilization may be fail when creation time
        Use re initialize function
        """
        try:
            self.channel = self._conf.pop('channel')
            pool = redis.ConnectionPool(**self._conf)
            self.conn = redis.Redis(connection_pool=pool)
            self.conn.ping()

            self.pubsub = self.conn.pubsub()
            self.pubsub.subscribe(self.channel)
            self.initialized = True

        except Exception:
            self.initialized = False

        return self.initialized


    ######################
    # FIFO Queue
    ######################
    def put(self, item):
        try:
            self.conn.rpush(self.channel, item)

        except redis.exceptions.ConnectionError as e:
            _LOGGER.error("####### Redis Queue put failed #############")
            _LOGGER.error(f"Redis connection error, reconnect after {WAIT_INTERVAL} sec ....")
            time.sleep(WAIT_INTERVAL)
            self.initialize()

        except Exception as e:
            _LOGGER.error("######## Contact to admin ##########")
            _LOGGER.error(f"Unknown error: {e}")


    def get(self):
        """
        blpop waits until item occurs
        """
        try:
            item = self.conn.blpop(self.channel)
            return item[1]
        except redis.exceptions.ConnectionError as e:
            _LOGGER.error("####### Redis Queue get failed #############")
            _LOGGER.error(f"Redis connection error, reconnect after {WAIT_INTERVAL} sec ....")
            time.sleep(WAIT_INTERVAL)
            self.initialize()

        except Exception as e:
            _LOGGER.error("######## Contact to admin ##########")
            _LOGGER.error(f"Unknown error: {e}")

    ######################
    # Pub / Sub
    ######################
    def subscribe(self):
        # This is redis.subscribe
        while True:
            m = self.pubsub.get_message()
            if m != None:
                return m
            time.sleep(0.01)

    def publish(self, item):
        # This is redis.publish
        try:
            return self.conn.publish(self.channel, item)
        except Exception as e:
            raise ERROR_UNKNOWN(message=e)



# -*- coding: utf-8 -*-
import logging
import json


class ErrorFilter(logging.Filter):
    def filter(self, record):
        if getattr(record, 'error_code', None) is None:
            record.error_code = ''

        if getattr(record, 'error_message', None) is None:
            record.error_message = ''

        if getattr(record, 'traceback', None) is None:
            record.traceback = ''
        else:
            record.traceback = json.dumps(record.traceback)

        return True



import threading
import traceback
import logging
from threading import local

from spaceone.core import utils, config
from opentelemetry import trace
from opentelemetry.trace import format_trace_id
from opentelemetry.trace.span import TraceFlags

__all__ = ['LOCAL_STORAGE', 'get_transaction', 'create_transaction', 'delete_transaction',
           'Transaction']

_LOGGER = logging.getLogger(__name__)
LOCAL_STORAGE = local()


class Transaction(object):

    def __init__(self, resource: str = None, verb: str = None, trace_id: str = None, meta=None):
        self._id = None
        self._thread_id = str(threading.current_thread().ident)
        self._service = config.get_service()
        self._resource = resource
        self._verb = verb
        self._rollbacks = []
        self._set_meta(meta)
        self._set_trace_id(trace_id)
        self._event_handlers = []

    def __repr__(self):
        return f"<Transaction ({self._resource}.{self._verb})>"

    def _set_trace_id(self, trace_id: str = None) -> None:
        if trace_id:
            self._id = trace_id
        else:
            self._id = format_trace_id(utils.generate_trace_id())

    def _set_meta(self, meta: dict = None):
        if meta:
            self._meta = meta.copy()
        else:
            self._meta = {}

    @property
    def id(self) -> str:
        return self._id

    @property
    def thread_id(self) -> str:
        return self._thread_id

    @property
    def service(self) -> str:
        return self._service

    @property
    def resource(self) -> str:
        return self._resource

    @property
    def verb(self) -> str:
        return self._verb

    def add_rollback(self, fn, *args, **kwargs) -> None:
        self._rollbacks.insert(0, {
            'fn': fn,
            'args': args,
            'kwargs': kwargs
        })

    def execute_rollback(self) -> None:
        for rollback in self._rollbacks:
            try:
                rollback['fn'](*rollback['args'], **rollback['kwargs'])
            except Exception:
                _LOGGER.info(f'[ROLLBACK-ERROR] {self}')
                _LOGGER.info(traceback.format_exc())

    @property
    def meta(self) -> dict:
        return self._meta

    def set_meta(self, key, value) -> None:
        self._meta[key] = value

    def get_meta(self, key, default=None):
        return self._meta.get(key, default)

    def get_connection_meta(self) -> list:
        keys = ['token', 'domain_id']
        result = []
        for key in keys:
            result.append((key, self.get_meta(key)))
        return result

    def notify_event(self, message):
        for handler in self._event_handlers:
            if not isinstance(message, dict):
                message = {'message': str(message)}

            handler.notify(self, 'IN_PROGRESS', message)


def get_transaction(is_create: bool = True) -> [Transaction, None]:
    current_span_context = trace.get_current_span().get_span_context()
    thread_id = str(threading.current_thread().ident)

    if current_span_context.trace_flags == TraceFlags.SAMPLED:
        trace_id_from_current_span = format_trace_id(current_span_context.trace_id)
        return getattr(LOCAL_STORAGE, trace_id_from_current_span, None)
    elif hasattr(LOCAL_STORAGE, thread_id):
        return getattr(LOCAL_STORAGE, thread_id, None)
    elif is_create:
        return create_transaction(thread_id=thread_id)
    else:
        return None


def create_transaction(resource: str = None, verb: str = None, trace_id: str = None,
                       meta: dict = None, thread_id: str = None) -> Transaction:
    transaction = Transaction(resource, verb, trace_id, meta)

    if thread_id:
        setattr(LOCAL_STORAGE, thread_id, transaction)
    else:
        setattr(LOCAL_STORAGE, transaction.id, transaction)

    return transaction


def delete_transaction() -> None:
    if transaction := get_transaction(is_create=False):
        if hasattr(LOCAL_STORAGE, transaction.id):
            delattr(LOCAL_STORAGE, transaction.id)

    thread_id = str(threading.current_thread().ident)
    if hasattr(LOCAL_STORAGE, thread_id):
        delattr(LOCAL_STORAGE, thread_id)



import sys
import contextlib
import functools
import inspect
import logging
import types
import copy
import traceback

from opentelemetry import trace, metrics
from opentelemetry.trace import NonRecordingSpan, format_trace_id
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from spaceone.core import config, utils
from spaceone.core.base import CoreObject
from spaceone.core.error import *
from spaceone.core.locator import Locator
from spaceone.core.transaction import Transaction, get_transaction, create_transaction, delete_transaction

from opentelemetry import trace

_LOGGER = logging.getLogger(__name__)
_TRACER = trace.get_tracer(__name__)


class BaseService(CoreObject):

    def __init__(self, metadata: dict = None, **kwargs):
        super().__init__(**kwargs)
        self.func_name = None
        # self.is_with_statement = False
        self.current_span_context = None
        self._metadata = metadata or {}

        self.handler = {
            'authentication': {'handlers': [], 'methods': []},
            'authorization': {'handlers': [], 'methods': []},
            'mutation': {'handlers': [], 'methods': []},
            'event': {'handlers': [], 'methods': []},
        }

        self.handler_exclude_apis = config.get_global('HANDLER_EXCLUDE_APIS', {})
        self.enable_stack_info = config.get_global('ENABLE_STACK_INFO', False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            raise exc_val


def transaction(func=None, verb=None, append_meta=None):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(self, params):
            _resource = self._metadata.get('resource') or self.__class__.__name__ or getattr(self, 'resource')
            _verb = verb or func.__name__
            with _TRACER.start_as_current_span(f'{_resource}.{_verb}',
                                               context=_get_span_context(self._metadata)) as span:
                self.current_span_context = span.get_span_context()
                trace_id = format_trace_id(self.current_span_context.trace_id)
                transaction = create_transaction(_resource, _verb, trace_id, self._metadata)
                return _pipeline(func, self, params, append_meta)

        return wrapped_func

    return wrapper(func) if func else wrapper


def _get_span_context(metadata):
    if traceparent := metadata.get('traceparent'):
        carrier = {'traceparent': traceparent}
        return TraceContextTextMapPropagator().extract(carrier)
    else:
        return None


def _pipeline(func, self, params, append_meta):
    try:
        self.func_name = func.__name__

        # 0. Set Extra Metadata
        if append_meta and isinstance(append_meta, dict):
            for key, value in append_meta.items():
                self.transaction.set_meta(key, value)
        with _TRACER.start_as_current_span('PreProcessing') as span:
            # 1. Start Event: Ignore exceptions
            if _check_handler_method(self, 'event'):
                for handler in self.handler['event']['handlers']:
                    try:
                        handler.notify(self.transaction, 'STARTED', params)
                    except Exception as e:
                        _LOGGER.error(f'{handler.__class__.__name__} Error (STARTED): {e}')

            # 2. Authentication:
            if _check_handler_method(self, 'authentication'):
                for handler in self.handler['authentication']['handlers']:
                    with _TRACER.start_as_current_span(handler.__class__.__name__) as span:
                        handler.verify(params)

            # 3. Authorization
            if _check_handler_method(self, 'authorization'):
                for handler in self.handler['authorization']['handlers']:
                    with _TRACER.start_as_current_span(handler.__class__.__name__) as span:
                        handler.verify(params)

            # 4. Print Info Log
            disable_info_log = str(self.transaction.get_meta('disable_info_log', 'false')).lower()
            if disable_info_log != 'true':
                _LOGGER.info('(REQUEST) =>', extra={'parameter': copy.deepcopy(params)})

            # 5. Mutation
            if _check_handler_method(self, 'mutation'):
                for handler in self.handler['mutation']['handlers']:
                    with _TRACER.start_as_current_span(handler.__class__.__name__) as span:
                        params = handler.request(params)

        # 6. Service Body
        with _TRACER.start_as_current_span(f'ServiceBody',
                                           links=[trace.Link(self.current_span_context)]) as span:
            response_or_iterator = func(self, params)

            # debug code for memory leak
            # local_storage = LOCAL_STORAGE.__dict__
            # _LOGGER.info(
            #     f'[BaseService] {get_transaction()} / number of items in local storage: {len(local_storage)} / items => {local_storage}')

            return response_or_iterator

        # # 7. Response Handlers
        # with _TRACER.start_as_current_span('PostProcessing') as span:
        #     if isinstance(response_or_iterator, types.GeneratorType):
        #         return _generate_response(self, response_or_iterator)
        #     else:
        #         response_or_iterator = _response_mutation_handler(self, response_or_iterator)
        #         _success_handler(self, response_or_iterator)
        #         return response_or_iterator

    except ERROR_INVALID_ARGUMENT as e:
        _error_handler(self, e)
        raise e

    except ERROR_BASE as e:
        _error_handler(self, e)
        raise e

    except Exception as e:
        error = ERROR_UNKNOWN(message=e)
        _error_handler(self, error)
        raise error

    finally:
        delete_transaction()


def _error_handler(self, error):
    if not isinstance(error, ERROR_BASE):
        error = ERROR_UNKNOWN(message=error)

    error.meta['skip_error_log'] = True

    # Failure Event
    if _check_handler_method(self, 'event'):
        for handler in self.handler['event']['handlers']:
            try:
                handler.notify('FAILURE', {'error_code': error.error_code, 'message': error.message})
            except Exception as e:
                _LOGGER.error(f'{handler.__class__.__name__} Error (FAILURE): {e}')

    _LOGGER.error(f'(Error) => {error.message} {error}', exc_info=True, stack_info=self.enable_stack_info)

    if transaction := get_transaction(is_create=False):
        transaction.execute_rollback()


def _success_handler(self, response):
    if _check_handler_method(self, 'event'):
        for handler in self.handler['event']['handlers']:
            try:
                with _TRACER.start_as_current_span(handler.__class__.__name__) as span:
                    handler.notify('SUCCESS', response)
            except Exception as e:
                _LOGGER.error(f'{handler.__class__.__name__} Error (SUCCESS): {e}')


def _response_mutation_handler(self, response):
    if _check_handler_method(self, 'mutation'):
        for handler in list(reversed(self.handler['mutation']['handlers'])):
            with _TRACER.start_as_current_span(handler.__class__.__name__) as span:
                response = handler.response(response)

    return response


def _generate_response(self, response_iterator):
    for response in response_iterator:
        response = _response_mutation_handler(self, response)
        _success_handler(self, response)
        yield response


def authentication_handler(func=None, methods='*', exclude=None):
    return _set_handler(func, 'authentication', methods, exclude)


def authorization_handler(func=None, methods='*', exclude=None):
    return _set_handler(func, 'authorization', methods, exclude)


def mutation_handler(func=None, methods='*', exclude=None):
    return _set_handler(func, 'mutation', methods, exclude)


def event_handler(func=None, methods='*', exclude=None):
    return _set_handler(func, 'event', methods, exclude)


def _set_handler(func, handler_type, methods, exclude=None):
    if exclude is None:
        exclude = []

    def wrapper(cls):
        @functools.wraps(cls)
        def wrapped_cls(*args, **kwargs):
            self = cls(*args, **kwargs)
            _load_handler(self, handler_type)
            return _bind_handler(self, handler_type, methods, exclude)

        return wrapped_cls

    if func:
        return wrapper(func)

    return wrapper


def _load_handler(self, handler_type):
    try:
        handlers = config.get_handler(handler_type)
        for handler in handlers:
            module_name, class_name = handler['backend'].rsplit('.', 1)
            handler_module = __import__(module_name, fromlist=[class_name])
            handler_conf = handler.copy()
            del handler_conf['backend']

            self.handler[handler_type]['handlers'].append(
                getattr(handler_module, class_name)(handler_conf))

    except ERROR_BASE as error:
        raise error

    except Exception as e:
        raise ERROR_HANDLER(handler_type=handler_type, reason=e)


def _get_service_methods(self):
    service_methods = []
    for f_name, f_object in inspect.getmembers(self.__class__, predicate=inspect.isfunction):
        if not f_name.startswith('__'):
            service_methods.append(f_name)

    return service_methods


def _bind_handler(self, handler_type, methods, exclude):
    handler_methods = []
    if methods == '*':
        handler_methods = _get_service_methods(self)
    else:
        if isinstance(methods, list):
            service_methods = _get_service_methods(self)
            for method in methods:
                if method in service_methods:
                    handler_methods.append(method)

    if isinstance(exclude, list):
        handler_methods = list(set(handler_methods) - set(exclude))

    self.handler[handler_type]['methods'] = \
        list(set(self.handler[handler_type]['methods'] + handler_methods))

    return self


def _check_handler_method(self, handler_type):
    if self.func_name in self.handler[handler_type]['methods']:
        tnx_method = f'{self.transaction.resource}.{self.transaction.verb}'
        exclude_apis = self.handler_exclude_apis.get(handler_type, [])

        if tnx_method in exclude_apis:
            return False
        else:
            return True
    else:
        return False



import re
import functools
from dateutil.parser import parse
from datetime import datetime
from dateutil.relativedelta import relativedelta

from spaceone.core import utils
from spaceone.core.error import *


def change_only_key(change_rule, key_path='only'):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            only_keys = change_rule.keys()
            only = utils.get_dict_value(params, key_path, [])
            new_only = []
            for key in only:
                key_match = key.split('.', 1)[0]
                if key_match in only_keys:
                    if change_rule[key_match] not in new_only:
                        new_only.append(change_rule[key_match])
                else:
                    new_only.append(key)

            utils.change_dict_value(params, key_path, new_only)
            return func(cls, params)

        return wrapped_func

    return wrapper


def check_required(required_keys):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            for key in required_keys:
                value = utils.get_dict_value(params, key)
                if value is None:
                    raise ERROR_REQUIRED_PARAMETER(key=key)

            return func(cls, params)

        return wrapped_func

    return wrapper


def set_query_page_limit(default_limit):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get('query', {})
            query['page'] = query.get('page', {})
            page_limit = query['page'].get('limit', 1000)

            if page_limit > default_limit:
                page_limit = default_limit

            query['page']['limit'] = page_limit
            params['query'] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def append_query_filter(filter_keys):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get('query', {})
            query['filter'] = query.get('filter', [])

            for key in filter_keys:
                if key in params:
                    if isinstance(params[key], list):
                        query['filter'].append({'k': key, 'v': params[key], 'o': 'in'})
                    else:
                        query['filter'].append({'k': key, 'v': params[key], 'o': 'eq'})

            params['query'] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def change_tag_filter(tag_key='tags'):
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            params['query'] = params.get('query', {})
            change_filter = []
            for condition in params['query'].get('filter', []):
                key = condition.get('key', condition.get('k'))
                if key and key.startswith(f'{tag_key}.'):
                    value = condition.get('value', condition.get('v'))
                    operator = condition.get('operator', condition.get('o'))
                    tag_key_split = key.split('.', 1)
                    change_filter.append({
                        'key': 'tags',
                        'value': {
                            'key': tag_key_split[1],
                            'value': _change_match_query(operator, value, condition)
                        },
                        'operator': 'match'
                    })
                else:
                    change_filter.append(condition)

            if 'only' in params['query']:
                change_only = []
                for key in params['query'].get('only', []):
                    if key.startswith('tags.') and 'tags' not in change_only:
                        change_only.append('tags')
                    else:
                        change_only.append(key)

                params['query']['only'] = change_only

            params['query']['filter'] = change_filter
            return func(cls, params)

        return wrapped_func

    return wrapper


def _change_match_query(operator, value, condition):
    if operator == 'eq':
        return value
    elif operator == 'not':
        return {
            '$ne': value
        }
    elif operator == 'in':
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {
            '$in': value
        }
    elif operator == 'not_in':
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {
            '$nin': value
        }
    elif operator == 'contain':
        return re.compile(value, re.IGNORECASE)
    elif operator == 'not_contain':
        return {
            '$not': re.compile(value, re.IGNORECASE)
        }
    elif operator == 'contain_in':
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {
            '$in': value
        }
    elif operator == 'not_contain_in':
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {
            '$nin': value
        }
    else:
        raise ERROR_DB_QUERY(reason='Filter operator is not supported.')


def append_keyword_filter(keywords=None):
    if keywords is None:
        keywords = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get('query', {})
            if 'keyword' in query:
                query['filter_or'] = query.get('filter_or', [])

                keyword = query['keyword'].strip()
                if len(keyword) > 0:
                    for key in keywords:
                        query['filter_or'].append({
                            'k': key,
                            'v': list(filter(None, keyword.split(' '))),
                            'o': 'contain_in'
                        })

                del query['keyword']
                params['query'] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def change_timestamp_value(timestamp_keys=None, timestamp_format='google_timestamp'):
    if timestamp_keys is None:
        timestamp_keys = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            change_params = {}

            for key, value in params.items():
                if key in timestamp_keys:
                    if not _is_null(value):
                        value = _convert_datetime_from_timestamp(value, key, timestamp_format)
                        change_params[key] = value
                else:
                    change_params[key] = value

            return func(cls, change_params)

        return wrapped_func

    return wrapper


def change_date_value(date_keys=None, date_format='%Y-%m-%d'):
    if date_keys is None:
        date_keys = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            change_params = {}

            for key, value in params.items():
                if key in date_keys:
                    if not _is_null(value):
                        value = _convert_date_from_string(value, key, date_format)
                        change_params[key] = value
                else:
                    change_params[key] = value

            return func(cls, change_params)

        return wrapped_func

    return wrapper


def change_timestamp_filter(filter_keys=None, timestamp_format='google_timestamp'):
    if filter_keys is None:
        filter_keys = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get('query', {})
            query_filter = query.get('filter')
            query_filter_or = query.get('filter_or')
            if query_filter:
                query['filter'] = _change_timestamp_condition(query_filter, filter_keys,
                                                              'filter', timestamp_format)

            if query_filter_or:
                query['filter_or'] = _change_timestamp_condition(query_filter_or, filter_keys,
                                                                 'filter_or', timestamp_format)

            params['query'] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def _is_null(value):
    if value is None or str(value).strip() == '':
        return True

    return False


def _change_timestamp_condition(query_filter, filter_keys, filter_type, timestamp_format):
    change_filter = []

    for condition in query_filter:
        key = condition.get('k') or condition.get('key')
        value = condition.get('v') or condition.get('value')
        operator = condition.get('o') or condition.get('operator')

        if key in filter_keys:
            value = _convert_datetime_from_timestamp(value, f'query.{filter_type}.{key}', timestamp_format)

        change_filter.append({
            'key': key,
            'value': value,
            'operator': operator
        })

    return change_filter


def _convert_datetime_from_timestamp(timestamp, key, timestamp_format):
    type_message = 'google.protobuf.Timestamp({seconds: <second>, nanos: <nano second>})'

    try:
        if timestamp_format == 'iso8601':
            type_message = 'ISO 8601(YYYY-MM-DDThh:mm:ssTZD)'
            return parse(timestamp)
        else:
            seconds = timestamp['seconds']
            nanos = timestamp.get('nanos')

            # TODO: change nanoseconds to timestamp

            return datetime.utcfromtimestamp((int(seconds)))
    except Exception as e:
        raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=type_message)


def _convert_date_from_string(date_str, key, date_format):
    try:
        return datetime.strptime(date_str, date_format).date()
    except Exception as e:
        raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_format)



from spaceone.api.sample.v1 import helloworld_pb2, helloworld_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class HelloWorld(BaseAPI, helloworld_pb2_grpc.HelloWorldServicer):

    pb2 = helloworld_pb2
    pb2_grpc = helloworld_pb2_grpc

    def say_hello(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('HelloWorldService', metadata) as helloworld_svc:
            return self.locator.get_info('HelloWorldInfo', helloworld_svc.say_hello(params))



# Service Configuration
PACKAGE = None
SERVICE = None

# Server Configuration
PORT = 50051
HOST = '127.0.0.1'

MAX_WORKERS = 100
MAX_MESSAGE_LENGTH = 1024 * 1024 * 1024

# Unit Test Configuration
MOCK_MODE = False

# gRPC Configuration
# gRPC Extension APIs
GRPC_EXTENSION_APIS = {
    'spaceone.core.extension.grpc_health': ['GRPCHealth'],
    'spaceone.core.extension.server_info': ['ServerInfo']
}

# REST Configuration
# REST Application Options
REST_TITLE = ''
REST_DESCRIPTION = ''
REST_CONTACT = {}

# REST Middlewares
REST_MIDDLEWARES = []

# REST Extension Routers
REST_EXTENSION_ROUTERS = [
    {
        'router_path': 'spaceone.core.fastapi.extension.health:router',
        'router_options': {}
    }
]

# REST Uvicorn Options
UVICORN_OPTIONS = {
    'factory': True
}


# Handler Configuration
HANDLERS = {
    'authentication': [],
    'authorization': [],
    'mutation': [],
    'event': []
}

HANDLER_EXCLUDE_APIS = {
    'authentication': [],
    'authorization': [],
    'mutation': [],
    'event': []
}

# Logging Configuration
ENABLE_STACK_INFO = False
LOG = {}

# OpenTelemetry Configuration
OTEL = {
    'endpoint': None
}

# Database Configuration
DATABASE_AUTO_CREATE_INDEX = True
DATABASE_NAME_PREFIX = ''
DATABASES = {
    'default': {}
}

# Cache Configuration
CACHES = {
    'default': {},
    'local': {
        'backend': 'spaceone.core.cache.local_cache.LocalCache',
        'max_size': 128,
        'ttl': 300
    }
}



import logging

from spaceone.core import pygrpc
from spaceone.core import utils
from spaceone.core.handler import BaseEventHandler
from spaceone.core.error import ERROR_HANDLER_CONFIGURATION

_LOGGER = logging.getLogger(__name__)


class EventGRPCHandler(BaseEventHandler):

    def __init__(self, config):
        super().__init__(config)
        self._initialize()

    def _initialize(self):
        if 'uri' not in self.config:
            _LOGGER.error(f'[_initialize] uri config is undefined.')
            raise ERROR_HANDLER_CONFIGURATION(handler='AuthenticationGRPCHandler')

        try:
            self.uri_info = utils.parse_grpc_uri(self.config['uri'])
        except Exception as e:
            _LOGGER.error(f'[_initialize] AuthenticationGRPCHandler Init Error: {e}')
            raise ERROR_HANDLER_CONFIGURATION(handler='AuthenticationGRPCHandler')

    def notify(self, status: str, message: dict):
        grpc_method = pygrpc.get_grpc_method(self.uri_info)
        grpc_method({
            'service': self.transaction.service,
            'resource': self.transaction.resource,
            'verb': self.transaction.verb,
            'status': status,
            'message': message
        })



class ERROR_BASE(Exception):
    _status_code = 'INTERNAL'
    _error_code = 'ERROR_BASE'
    _message = 'Base Error Class'
    _meta = {}

    def __init__(self, _error_code=None, _meta=None, **kwargs):
        if isinstance(_meta, dict):
            self._meta = _meta

        if _error_code:
            self._error_code = _error_code.strip()
        else:
            self._error_code = self.__class__.__name__

        try:
            self._message = self.message.format(**kwargs)
        except Exception:
            raise ERROR_MESSAGE_FORMAT(error_class=self.__class__.__name__, key=str(kwargs))

    @property
    def message(self):
        return self._message

    @property
    def error_code(self):
        return self._error_code

    @error_code.setter
    def error_code(self, value):
        self._error_code = value

    @property
    def status_code(self):
        return self._status_code

    @status_code.setter
    def status_code(self, value):
        self._status_code = value

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, key, value):
        self._meta[key] = value

    def _repr(self):
        return ('\n'
                f'\terror_code = {self._error_code}\n'
                f'\tstatus_code = {self._status_code}\n'
                f'\tmessage = {self._message}')

    def __repr__(self):
        return self._repr()

    def __str__(self):
        return self._repr()


class ERROR_INVALID_ARGUMENT(ERROR_BASE):
    _status_code = 'INVALID_ARGUMENT'
    _message = 'Argument is invalid.'


class ERROR_REQUIRED_PARAMETER(ERROR_INVALID_ARGUMENT):
    _message = 'Required parameter. (key = {key})'


class ERROR_INVALID_PARAMETER_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'Parameter type is invalid. (key = {key}, type = {type})'


class ERROR_INVALID_PARAMETER(ERROR_INVALID_ARGUMENT):
    _message = 'Parameter is invalid. (key = {key}, reason = {reason})'


class ERROR_NOT_FOUND(ERROR_INVALID_ARGUMENT):
    _status_code = 'NOT_FOUND'
    _message = 'Value not found. ({key} = {value})'


class ERROR_NOT_UNIQUE(ERROR_INVALID_ARGUMENT):
    _status_code = 'ALREADY_EXISTS'
    _message = 'Tried to save duplicate unique key. ({key} = {value})'


class ERROR_SAVE_UNIQUE_VALUES(ERROR_NOT_UNIQUE):
    _message = 'Tried to save duplicate unique values. (keys = {keys})'


class ERROR_EXIST_RESOURCE(ERROR_INVALID_ARGUMENT):
    _message = '\'{child}\' resources is existed in {parent}.'


class ERROR_OPERATOR_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'The value of \'{operator} operator\' does not support list type. ({condition})'


class ERROR_OPERATOR_LIST_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'The value of \'{operator} operator\' must be a list type. ({condition})'


class ERROR_OPERATOR_DICT_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'The value of \'{operator} operator\' must be a dictionary type. ({condition})'


class ERROR_OPERATOR_BOOLEAN_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'The value of \'{operator} operator\' must be a boolean type. ({condition})'


class ERROR_INVALID_FORMAT(ERROR_INVALID_ARGUMENT):
    _message = 'Value format is invalid. ({key} = {value} ! {rule})'


class ERROR_JSON_FORMAT(ERROR_INVALID_ARGUMENT):
    _message = 'JSON format is invalid. ({key} = {value})'


class ERROR_AUTHENTICATE_FAILURE(ERROR_INVALID_ARGUMENT):
    _status_code = 'UNAUTHENTICATED'
    _message = 'Authenticate failure. (message = {message})'


class ERROR_PERMISSION_DENIED(ERROR_INVALID_ARGUMENT):
    _status_code = 'PERMISSION_DENIED'
    _message = 'Permission denied.'


class ERROR_UNKNOWN(ERROR_BASE):
    _status_code = 'INTERNAL'
    _message = '{message}'


class ERROR_REQUEST_TIMEOUT(ERROR_UNKNOWN):
    _status_code = 'DEADLINE_EXCEEDED'
    _message = 'Request timeout!'


class ERROR_TRANSACTION_STATUS(ERROR_UNKNOWN):
    _message = 'Transaction status is incorrect. (status = {status})'


class ERROR_UNSUPPORTED_API(ERROR_UNKNOWN):
    _message = 'Call Unsupported API. (reason = {reason})'


class ERROR_CONFIGURATION(ERROR_UNKNOWN):
    _message = 'Configuration is invalid. (key = {key})'


class ERROR_DB_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'Database configuration is invalid. (backend = {backend})'


class ERROR_CACHE_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'Cache configuration is invalid. (backend = {backend})'


class ERROR_CONNECTOR_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'Connector configuration is invalid. (backend = {backend})'


class ERROR_GRPC_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'gRPC client configuration is invalid. ({endpoint}/{service}/{method})'


class ERROR_HANDLER_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'Handler configuration is invalid. (handler = {handler})'


class ERROR_CONNECTOR_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'Connector configuration is invalid. (connector = {connector})'


class ERROR_LOG_CONFIG(ERROR_CONFIGURATION):
    _message = 'Log configuration is invalid. (reason = {reason})'


class ERROR_WRONG_CONFIGURATION(ERROR_CONFIGURATION):
    _message = 'Configuration is invalid. ({key})'


class ERROR_DB_QUERY(ERROR_UNKNOWN):
    _message = 'Database query failed. (reason = {reason})'


class ERROR_CACHE_OPTION(ERROR_UNKNOWN):
    _message = 'Does not support the cache option. (method = {method}, option = {option})'


class ERROR_CACHE_TIMEOUT(ERROR_UNKNOWN):
    _message = 'Cache timeout error. (config = {config})'


class ERROR_CACHE_ENCODE(ERROR_UNKNOWN):
    _message = 'Cache data encoding failed. (reason = {reason})'


class ERROR_CACHE_DECODE(ERROR_UNKNOWN):
    _message = 'Cache data decoding failed. (reason = {reason})'


class ERROR_CACHEABLE_VALUE_TYPE(ERROR_UNKNOWN):
    _message = 'The value of cache.cacheable must be a dict type.'


class ERROR_INTERNAL_API(ERROR_UNKNOWN):
    _message = '{message}'


class ERROR_UNAVAILAVBLE(ERROR_UNKNOWN):
    _status_code = 'UNAVAILABLE'
    _message = 'Server is unavailable. (reason = {reason})'


class ERROR_QUEUE_PUT(ERROR_UNAVAILAVBLE):
    _message = 'Queue data put failed. (reason = {reason})'


class ERROR_QUEUE_GET(ERROR_UNAVAILAVBLE):
    _message = 'Queue data get failed. (reason = {reason})'


class ERROR_GRPC_CONNECTION(ERROR_UNAVAILAVBLE):
    _status_code = 'UNAVAILABLE'
    _message = 'Server is unavailable. (channel = {channel}, message = {message})'


class ERROR_GRPC_TLS_HANDSHAKE(ERROR_GRPC_CONNECTION):
    _message = 'TLS handshake failed. (reason = {reason})'


class ERROR_HANDLER(ERROR_UNKNOWN):
    _message = '\'{handler_type} handler\' import failed. (reason = {reason})'


class ERROR_CONNECTOR(ERROR_UNKNOWN):
    _message = '\'{connector} handler\' import failed. (reason = {reason})'


class ERROR_CONNECTOR_LOAD(ERROR_CONNECTOR):
    _message = 'Failed to load connector. (connector = {connector}, reason = {reason})'


class ERROR_LOCATOR(ERROR_UNKNOWN):
    _message = '\'{name}\' load failed. (reason = {reason})'


class ERROR_TASK_LOCATOR(ERROR_LOCATOR):
    _message = 'Call locator failure. locator: {locator}, name: {name}'


class ERROR_TASK_METHOD(ERROR_UNKNOWN):
    _message = 'Call method failure. name: {name}, method: {method}, params: {params}'


class ERROR_NOT_IMPLEMENTED(ERROR_UNKNOWN):
    _status_code = 'UNIMPLEMENTED'
    _message = 'Not implemented, {name}'


class ERROR_MESSAGE_FORMAT(ERROR_UNKNOWN):
    _message = 'Error message format is invalid. (error_class={error_class}, key={key})'


class ERROR_CACHE_KEY_FORMAT(ERROR_UNKNOWN):
    _message = 'Cache key format is invalid. (key={key})'



from spaceone.core.service import *

__all__ = ['HelloWorldService']


@authentication_handler
@authorization_handler
@event_handler
class HelloWorldService(BaseService):

    @transaction
    @check_required(['name'])
    def say_hello(self, params):
        helloworld_mgr = self.locator.get_manager('HelloWorldManager')
        return helloworld_mgr.say_hello(params['name'])



import os
import re
import time
import random
import string
import secrets
import sys
import datetime
import yaml
import json
import hashlib
import urllib
import pkg_resources

from urllib.parse import urlparse
from dateutil.parser import isoparse
from typing import Tuple
from pathlib import Path
from typing import Union

YAML_LOADER = yaml.Loader
YAML_LOADER.add_implicit_resolver(
    u'tag:yaml.org,2002:float',
    re.compile(u'''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$''', re.X),
    list(u'-+0123456789.'))


def generate_id(prefix: str = 'id', nbytes: int = 6) -> str:
    random_id = secrets.token_hex(nbytes)
    return f'{prefix}-{random_id}'


def generate_trace_id() -> int:
    return random.getrandbits(128)


def generate_secret(nbytes: int = 32) -> str:
    return secrets.token_urlsafe(nbytes)


def generate_password(length: int = 8) -> str:
    # create alphanumerical from string constants
    printable = f'{string.ascii_letters}{string.digits}'

    # convert printable from string to list and shuffle
    printable = list(printable)
    random.shuffle(printable)

    # generate random password and convert to string
    random_password = random.choices(printable, k=length)

    # Append one lowercase and one uppercase and one number characters
    random_password.append(random.choices(list(string.ascii_lowercase))[0])
    random_password.append(random.choices(list(string.ascii_uppercase))[0])
    random_password.append(random.choices(list(string.digits))[0])

    random_password = ''.join(random_password)
    return random_password


def random_string(nbytes: int = 6) -> str:
    return secrets.token_hex(nbytes)


def create_dir(path: str):
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)


def dump_json(data, indent=None, sort_keys=False) -> str:
    try:
        return json.dumps(data, indent=indent, sort_keys=sort_keys, ensure_ascii=False)
    except Exception as e:
        raise ValueError(f'JSON Dump Error: {str(e)}')


def save_json_to_file(data, json_file: str, indent=None, sort_keys=False):
    try:
        json_str = dump_json(data, indent, sort_keys)
        with open(json_file, 'w') as f:
            f.write(json_str)
            f.close()
    except Exception as e:
        raise ValueError(f'JSON Save Error: {str(e)}')


def load_json(json_str: str) -> dict:
    try:
        return json.loads(json_str)
    except Exception:
        raise ValueError(f'JSON Load Error: {json_str}')


def load_json_from_file(json_file: str) -> dict:
    try:
        with open(json_file, 'r') as f:
            return load_json(f.read())
    except Exception:
        raise Exception(f'JSON Load Error: {json_file}')


def dump_yaml(data: dict) -> str:
    try:
        return yaml.dump(data, allow_unicode=True)
    except Exception as e:
        raise ValueError(f'YAML Dump Error: {str(e)}')


def save_yaml_to_file(data: dict, yaml_file: str):
    try:
        yaml_str = dump_yaml(data)
        with open(yaml_file, 'w') as f:
            f.write(yaml_str)
            f.close()
    except Exception as e:
        raise ValueError(f'YAML Save Error: {str(e)}')


def load_yaml(yaml_str: str) -> dict:
    try:
        return yaml.load(yaml_str, Loader=YAML_LOADER)
    except Exception:
        raise ValueError(f'YAML Load Error: {yaml_str}')


def load_yaml_from_file(yaml_file: str) -> dict:
    try:
        with open(yaml_file, 'r') as f:
            return load_yaml(f.read())
    except Exception:
        raise Exception(f'YAML Load Error: {yaml_file}')


def load_yaml_from_url(url: str) -> dict:
    try:
        f = urllib.urlopen(url)
        return load_yaml(f.read())
    except Exception as e:
        raise Exception(f'Http call error: {url}. e={e}')


def parse_endpoint(endpoint: str) -> dict:
    try:
        o = urlparse(endpoint)
    except Exception:
        raise ValueError(f'Endpoint is invalid. ({endpoint})')

    return {
        'scheme': o.scheme,
        'hostname': o.hostname,
        'port': o.port,
        'path': o.path
    }


def parse_grpc_endpoint(endpoint: str) -> dict:
    try:
        endpoint_info = parse_endpoint(endpoint)

        if endpoint_info['scheme'] not in ['grpc', 'grpc+ssl']:
            raise ValueError(f'Unsupported protocol. (supported_protocol=grpc|grpc+ssl, endpoint={endpoint})')

    except Exception:
        raise ValueError(f'gRPC Endpoint is invalid. (endpoint={endpoint})')

    if endpoint_info['scheme'] == 'grpc+ssl':
        ssl_enabled = True
    else:
        ssl_enabled = False

    return {
        'endpoint': f'{endpoint_info["hostname"]}:{endpoint_info["port"]}',
        'ssl_enabled': ssl_enabled
    }


def parse_grpc_uri(uri: str) -> dict:
    try:
        endpoint_info = parse_endpoint(uri)

        if endpoint_info['scheme'] not in ['grpc', 'grpc+ssl']:
            raise ValueError(f'Unsupported protocol. (supported_protocol=grpc|grpc+ssl, uri={uri})')

        version, service, method = \
            filter(lambda x: x.strip() != '', endpoint_info['path'].split('/'))
    except Exception:
        raise ValueError(f'gRPC URI is invalid. (uri={uri})')

    if endpoint_info['scheme'] == 'grpc+ssl':
        ssl_enabled = True
    else:
        ssl_enabled = False

    return {
        'endpoint': f'{endpoint_info["hostname"]}:{endpoint_info["port"]}',
        'version': version,
        'service': service,
        'method': method,
        'ssl_enabled': ssl_enabled
    }


def deep_merge(from_dict: dict, into_dict: dict) -> dict:
    for key, value in from_dict.items():
        if isinstance(value, dict):
            node = into_dict.setdefault(key, {})
            deep_merge(value, node)
        else:
            into_dict[key] = value

    return into_dict


def parse_timediff_query(query: str) -> datetime:
    try:
        time_info, include_operator = _parse_timediff_from_regex(query)
        base_dt = _convert_base_time(time_info['base_time'])
        if include_operator:
            time_delta = _convert_time_delta(time_info['time_delta_number'], time_info['time_delta_unit'])
            if time_info['operator'] == '+':
                return base_dt + time_delta
            else:
                return base_dt - time_delta
        else:
            return base_dt
    except Exception as e:
        raise ValueError(f'Timediff format is invalid. (value={query})')


def _convert_time_delta(time_delta_number: str, time_delta_unit: str) -> datetime.timedelta:
    _time_delta_map = {
        's': 'seconds',
        'm': 'minutes',
        'h': 'hours',
        'd': 'days',
        'w': 'weeks'
    }

    time_delta_params = {
        _time_delta_map[time_delta_unit]: int(time_delta_number)
    }

    return datetime.timedelta(**time_delta_params)


def _convert_base_time(time_str: str) -> datetime:
    now = datetime.datetime.utcnow()
    if time_str == 'now':
        return now
    elif time_str == 'now/d':
        return datetime.datetime.combine(now.date(), datetime.time(0))
    elif time_str == 'now/w':
        today = datetime.datetime.combine(now.date(), datetime.time(0))
        return today - datetime.timedelta(days=now.date().weekday())
    elif time_str == 'now/m':
        today = datetime.datetime.combine(now.date(), datetime.time(0))
        return today.replace(day=1)


def _parse_timediff_from_regex(query: str) -> Tuple[dict, bool]:
    p = r'^\s?(?P<base_time>now(\/[dwm])?)\s?' \
        r'(?P<operator>[+|-])\s?' \
        r'(?P<time_delta_number>\d+)' \
        r'(?P<time_delta_unit>[s|m|h|d|w])\s?$'
    rule = re.compile(p)
    match = rule.match(query)
    if match:
        return match.groupdict(), True
    else:
        if query.strip() not in ['now', 'now/d', 'now/w', 'now/m']:
            raise ValueError(f'Timediff format is invalid. (value={query})')

        return {'base_time': query.strip()}, False


def get_dict_value(data: dict, dotted_key: str, default_value: any = None):
    if '.' in dotted_key:
        key, rest = dotted_key.split(".", 1)

        if isinstance(data, dict) and key in data:
            if isinstance(data[key], list):
                return get_list_values(data[key], rest, default_value)
            else:
                return get_dict_value(data[key], rest, default_value)

        else:
            return default_value
    else:
        if isinstance(data, dict):
            return data.get(dotted_key, default_value)
        else:
            return default_value


def get_list_values(values: list, dotted_key: str, default_value: any = None):
    match_option = 'contain'
    if len(dotted_key) > 1 and dotted_key[0] == "?":
        condition = True
        try:
            cond_option, rest = dotted_key[1:].split('=>', 1)
            cond_key, cond_value = cond_option.split(':', 1)
        except Exception as e:
            # Syntax Error
            return default_value

        if cond_value[:1] == '=':
            match_option = 'eq'
            cond_value = cond_value[1:]

        elif cond_value[:1] == '!':
            match_option = 'not'
            cond_value = cond_value[1:]

    else:
        try:
            # Get value by index
            if '.' in dotted_key:
                index, rest = dotted_key.split('.', 1)
                index = int(index)
                if index >= len(values):
                    return default_value
                else:
                    if rest.strip() != '' and isinstance(values[index], dict):
                        return get_dict_value(values[index], rest, default_value)
                    else:
                        values = values[index]
            else:
                return values[int(dotted_key)]

        except Exception:
            rest = dotted_key

        condition = False
        cond_key = None
        cond_value = None

    results = []
    for value in values:
        # Get value from condition
        if not isinstance(value, dict) \
                or (condition and not _check_condition(match_option, value[cond_key], cond_value)):
            continue

        # Get value from dict key
        result = get_dict_value(value, rest, default_value)

        if result:
            if isinstance(result, list):
                results += result
            else:
                results.append(result)

    try:
        if len(results) > 0:
            return list(set(results))
        else:
            return default_value
    except Exception:
        return results


def _check_condition(match_option: str, val1, val2):
    val1 = str(val1).lower()
    val2 = str(val2).lower()

    if match_option == 'eq':
        if val1 == val2:
            return True

    elif match_option == 'not':
        if val1.find(val2) < 0:
            return True

    else:
        if val1.find(val2) >= 0:
            return True

    return False


def change_dict_value(data: dict, dotted_key: str, change_value, change_type='value') -> dict:
    # change_value = func or value(any type)
    if '.' in dotted_key:
        key, rest = dotted_key.split('.', 1)
        if key in data:
            if rest.startswith('[]') and isinstance(data[key], list):
                list_data = []
                for sub_data in data[key]:
                    if rest.strip() == '[]':
                        list_data.append(_change_value_by_type(change_type, sub_data, change_value))
                    else:
                        sub_rest = rest.split('.', 1)[1]
                        list_data.append(change_dict_value(sub_data, sub_rest, change_value, change_type))
                data[key] = list_data
            elif isinstance(data[key], dict):
                data[key] = change_dict_value(data[key], rest, change_value, change_type)
    else:
        if dotted_key in data:
            data[dotted_key] = _change_value_by_type(change_type, data[dotted_key], change_value)

    return data


def _change_value_by_type(change_type, original_value, change_value):
    if change_type == 'value':
        return change_value
    elif change_type == 'func':
        return change_value(original_value)
    else:
        return original_value


def date_to_string(value: datetime.date, date_format: str = '%Y-%m-%d') -> Union[str, None]:
    if isinstance(value, datetime.date):
        return value.strftime(date_format)

    return None


def datetime_to_iso8601(value: datetime.datetime) -> Union[str, None]:
    if isinstance(value, datetime.datetime):
        value = value.replace(tzinfo=None)
        return f"{value.isoformat(timespec='milliseconds')}Z"

    return None


def iso8601_to_datetime(value: str) -> Union[datetime.datetime, None]:
    if isinstance(value, str):
        try:
            return isoparse(value)
        except Exception as e:
            raise ValueError(f'Datetime(ISO8601) format is invalid. (value={value})')

    return None


def iso8601_to_timestamp(value: str) -> Union[int, None]:
    dt = iso8601_to_datetime(value)
    if dt:
        return int(time.mktime(dt.timetuple()))

    return None


def tags_to_dict(tags: list) -> Union[dict, None]:
    if isinstance(tags, list):
        dict_value = {}
        for tag in tags:
            dict_value[tag['key']] = tag['value']

        return dict_value

    else:
        return None


def dict_to_tags(dict_value: dict) -> list:
    tags = []
    if isinstance(dict_value, dict):
        for key, value in dict_value.items():
            tags.append({
                'key': key,
                'value': value
            })

    return tags


def dict_to_hash(dict_value: dict) -> str:
    """MD5 hash of a dictionary."""
    dhash = hashlib.md5()
    encoded = json.dumps(dict_value, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def string_to_hash(str_value: str) -> str:
    """MD5 hash of a String."""
    dhash = hashlib.md5(str_value.encode("utf-8"))
    return dhash.hexdigest()


def change_dict_with_dot_notation(dict_value: dict, key='', dots=None) -> dict:
    if dots is None:
        dots = {}
    if isinstance(dict_value, dict):
        for (k, v) in dict_value.items():
            change_dict_with_dot_notation(dict_value[k], f'{key}.{k}' if key else k, dots)
    else:
        dots[key] = dict_value
    return dots


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
import logging


class TracebackFilter(logging.Filter):
    def filter(self, record):
        _traceback = getattr(record, 'traceback', None)

        if _traceback is None or _traceback == "":
            record.traceback = ''
        else:
            record.traceback = _traceback.replace('\\n', '\n').replace('\\"', '\"').replace('\\t', '\t')

        return True


class TracebackLogFilter(logging.Filter):
    def filter(self, record):
        _traceback = getattr(record, 'traceback', None)

        if _traceback is None or _traceback == "":
            record.traceback_log = '""'
        else:
            record.traceback_log = _traceback

        return True


import os
import shutil
import sys
import unittest
from typing import List

import click
import pkg_resources

from spaceone.core import config, pygrpc, fastapi, utils
from spaceone.core import scheduler as scheduler_v1
from spaceone.core.unittest.runner import RichTestRunner


@click.group()
def cli():
    pass


@cli.command()
@click.argument('project_name')
@click.option('-d', '--directory', type=click.Path(), help='Project directory')
def create_project(project_name=None, directory=None):
    _create_project(project_name, directory)


@cli.command()
@click.argument('package')
@click.option('-p', '--port', type=int, default=lambda: os.environ.get('SPACEONE_PORT', 50051),
              help='Port of gRPC server', show_default=True)
@click.option('-c', '--config-file', type=click.Path(exists=True), default=lambda: os.environ.get('SPACEONE_CONFIG_FILE'),
              help='Config file path')
@click.option('-m', '--module-path', type=click.Path(exists=True), multiple=True, help='Module path')
def grpc(package, port=None, config_file=None, module_path=None):
    """Run a gRPC server"""
    _set_server_config(package, module_path, port, config_file=config_file)
    pygrpc.serve()


@cli.command()
@click.argument('package')
@click.option('-h', '--host', type=str, default=lambda: os.environ.get('SPACEONE_HOST', '127.0.0.1'),
              help='Host of REST server', show_default=True)
@click.option('-p', '--port', type=int, default=lambda: os.environ.get('SPACEONE_PORT', 8000),
              help='Port of REST server', show_default=True)
@click.option('-c', '--config-file', type=click.Path(exists=True), default=lambda: os.environ.get('SPACEONE_CONFIG_FILE'),
              help='Config file path')
@click.option('-m', '--module-path', type=click.Path(exists=True), multiple=True, help='Module path')
def rest(package, host=None, port=None, config_file=None, module_path=None):
    """Run a FastAPI REST server"""
    _set_server_config(package, module_path, port, config_file=config_file)
    fastapi.serve()


@cli.command()
@click.argument('package')
@click.option('-c', '--config-file', type=click.Path(exists=True), default=lambda: os.environ.get('SPACEONE_CONFIG_FILE'),
              help='config file path')
@click.option('-m', '--module-path', type=click.Path(exists=True), multiple=True, help='Module path')
def scheduler(package, config_file=None, module_path=None):
    """Run a scheduler server"""
    _set_server_config(package, module_path, config_file=config_file)
    scheduler_v1.serve()


@cli.command()
@click.argument('package')
@click.option('-c', '--config-file', type=click.Path(exists=True), default=lambda: os.environ.get('SPACEONE_CONFIG_FILE'),
              help='Config file path')
@click.option('-m', '--module-path', type=click.Path(exists=True), multiple=True, help='Module path')
@click.option('-o', '--output', default='yaml', help='Output format',
              type=click.Choice(['json', 'yaml']), show_default=True)
def show_config(package, config_file=None, module_path=None, output=None):
    """Show global configurations"""
    _set_server_config(package, module_path, config_file=config_file)
    _print_config(output)


@cli.command()
@click.option('-c', '--config-file', type=str, help='Config file path')
@click.option('-d', '--dir', type=str, help="Directory containing test files",
              default=lambda: os.environ.get('SPACEONE_WORKING_DIR', os.getcwd()))
@click.option('-f', '--failfast', help="Fast failure flag", is_flag=True)
@click.option('-s', '--scenario', type=str, help="Scenario file path")
@click.option('-p', '--parameters', type=str, help="Custom parameters to override a scenario file. "
                                                   "ex) -p domain.domain.name=new_name -p options.update_mode=false",
              multiple=True)
@click.option('-v', '--verbose', count=True, help='Verbosity level', default=1)
def test(config_file=None, dir=None, failfast=False, scenario: str = None, parameters: List[str] = None, verbose=1):
    """Unit tests for source code"""
    # set config
    if config:
        os.environ['TEST_CONFIG'] = config_file

    if scenario:
        os.environ['TEST_SCENARIO'] = scenario

    if parameters:
        os.environ['TEST_SCENARIO_PARAMS'] = ','.join(parameters)

    # run test
    loader = unittest.TestLoader()
    suites = loader.discover(dir)

    full_suite = unittest.TestSuite()
    full_suite.addTests(suites)
    RichTestRunner(verbosity=verbose, failfast=failfast).run(full_suite)


def _set_python_path(package, module_path):
    current_path = os.getcwd()

    if current_path not in sys.path:
        sys.path.insert(0, current_path)

    if isinstance(module_path, tuple):
        for path in module_path:
            if path not in sys.path:
                sys.path.insert(0, path)

    if '.' in package:
        pkg_resources.declare_namespace(package)

    try:
        __import__(package)
    except Exception:
        raise Exception(f'The package({package}) can not imported. '
                        'Please check the module path.')


def _set_server_config(package, module_path=None, port=None, config_file=None):
    # 1. Set a python path
    _set_python_path(package, module_path)

    # 2. Initialize config from command argument
    config.init_conf(
        package=package,
        port=port
    )

    # 3. Get service config from global_conf.py
    config.set_service_config()

    # 4. Merge file conf
    if config_file:
        config.set_file_conf(config_file)


def init_project_file(path, text):
    with open(path, 'w') as f:
        f.write(text)


def _create_project(project_name, directory=None):
    # Initialize path
    project_name = project_name
    project_directory = directory or os.getcwd()
    project_path = os.path.join(project_directory, project_name)

    # Copy skeleton source code
    skeleton_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'skeleton')
    shutil.copytree(skeleton_path, project_path, ignore=shutil.ignore_patterns('__pycache__'))

    # Change source code for new project environment
    init_project_file(os.path.join(project_path, 'service', '__init__.py'),
                      f'from {project_name}.service.helloworld_service import *\n')
    init_project_file(os.path.join(project_path, 'manager', '__init__.py'),
                      f'from {project_name}.manager.helloworld_manager import *\n')
    init_project_file(os.path.join(project_path, 'info', '__init__.py'),
                      f'from {project_name}.info.helloworld_info import *\n')
    proto_conf = (
        "PROTO = {\n"
        f"    '{project_name}.api.helloworld': ['HelloWorld']\n"
        "}\n"
    )
    init_project_file(os.path.join(project_path, 'conf', 'proto_conf.py'), proto_conf)


def _print_config(output):
    data = {
        'GLOBAL': config.get_global()
    }

    if output == 'json':
        print(utils.dump_json(data, indent=4))
    else:
        print(utils.dump_yaml(data))


if __name__ == '__main__':
    cli()



import logging
from spaceone.core.handler import BaseMutationHandler

_LOGGER = logging.getLogger(__name__)


class SpaceONEMutationHandler(BaseMutationHandler):

    def request(self, params):
        scope = self.transaction.get_meta('authorization.scope')
        role_type = self.transaction.get_meta('authorization.role_type')

        domain_id = self.transaction.get_meta('domain_id') or params.get('domain_id')

        if role_type and role_type != 'SYSTEM':
            if scope == 'DOMAIN':
                params = self._apply_domain_scope(params, role_type, domain_id)
            elif scope == 'PROJECT':
                params = self._apply_project_scope(params, role_type, domain_id)
            elif scope == 'USER':
                params = self._apply_user_scope(params, role_type, domain_id)
            elif scope == 'DOMAIN_OR_PROJECT':
                params = self._apply_domain_or_project_scope(params, role_type, domain_id)
            elif scope == 'PUBLIC_OR_DOMAIN':
                params = self._apply_public_or_domain_scope(params, role_type, domain_id)

        return params

    def _apply_domain_scope(self, params, role_type, domain_id):
        params['domain_id'] = domain_id
        return params

    def _apply_project_scope(self, params, role_type, domain_id):
        params['domain_id'] = domain_id

        if role_type == 'PROJECT':
            params['user_projects'] = self.transaction.get_meta('authorization.projects')
            params['user_project_groups'] = self.transaction.get_meta('authorization.project_groups')

        return params

    def _apply_user_scope(self, params, role_type, domain_id):
        params['domain_id'] = domain_id

        if role_type in ['PROJECT', 'USER']:
            params['user_id'] = self.transaction.get_meta('user_id')

        return params

    def _apply_domain_or_project_scope(self, params, role_type, domain_id):
        params['domain_id'] = domain_id

        if role_type == 'PROJECT':
            params['user_projects'] = self.transaction.get_meta('authorization.projects') + [None]
            params['user_project_groups'] = self.transaction.get_meta('authorization.project_groups') + [None]

        return params

    def _apply_public_or_domain_scope(self, params, role_type, domain_id):
        params['user_domains'] = [domain_id, None]

        return params



import abc
from typing import Any
from spaceone.core.base import CoreObject


class BaseHandler(CoreObject):

    def __init__(self, config: dict):
        super().__init__()

        self.config = config


class BaseAuthenticationHandler(abc.ABC, BaseHandler):

    @abc.abstractmethod
    def verify(self, params: dict) -> None:
        """
        Args:
            params (dict): Request Parameter

        Returns:
            None
        """
        raise NotImplementedError('verify method not implemented!')


class BaseAuthorizationHandler(abc.ABC, BaseHandler):

    @abc.abstractmethod
    def verify(self, params: dict) -> None:
        """
        Args:
            params (dict): Request Parameter

        Returns:
            None
        """
        raise NotImplementedError('verify method not implemented!')


class BaseMutationHandler(BaseHandler):

    def request(self, params) -> Any:
        """
        Args:
            params (dict): Request Parameter

        Returns:
            params (dict): Changed Parameter
        """
        return params

    def response(self, result: Any) -> Any:
        """
        Args:
            result (dict): Result data

        Returns:
            result (dict): Changed Result Data
        """
        return result


class BaseEventHandler(abc.ABC, BaseHandler):

    @abc.abstractmethod
    def notify(self, status: str, message: dict) -> None:
        """
        Args:
            status (str): 'Transaction Status: STARTED or IN_PROGRESS or SUCCESS or FAILURE'
            message (dict): 'Request Parameter or Response Data or Error Message'

        Returns:
            None
        """
        raise NotImplementedError('notify method not implemented!')



import logging
from spaceone.core.transaction import get_transaction


class TransactionFilter(logging.Filter):

    def filter(self, record):
        if transaction := get_transaction(is_create=False):
            record.service = transaction.service
            record.trace_id = transaction.id
            record.domain_id = transaction.get_meta('domain_id', '')
            record.user_id = transaction.get_meta('user_id', '')
            record.peer = transaction.get_meta('peer')
            if transaction.resource and transaction.verb:
                record.tnx_method = f'{transaction.resource}.{transaction.verb}'
            else:
                record.tnx_method = ''
        else:
            record.service = ""
            record.trace_id = ""
            record.domain_id = ""
            record.user_id = ""
            record.tnx_status = ""
            record.tnx_method = ""
            record.peer = ""
        return True
